"""
Package that contains base classes for solvers

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

import logging
from typing import Dict, Any, List  # @UnusedImport
from abc import ABCMeta, abstractmethod

import numba as nb

from ..pdes.base import PDEBase
from ..fields.base import FieldBase
from ..tools.misc import classproperty



class SolverBase(metaclass=ABCMeta):
    """ base class for simulations """

    
    _subclasses: Dict[str, 'SolverBase'] = {}  # all inheriting classes
    
        
    def __init__(self, pde: PDEBase):
        """ initialize the solver
        
        Args:
            pde (:class:`~pde.pdes.base.PDEBase`):
                The partial differential equation that should be solved
        """
        self.pde = pde
        self.info: Dict[str, Any] = {'class': self.__class__.__name__,
                                     'pde_class': self.pde.__class__.__name__}
        self._logger = logging.getLogger(self.__class__.__name__)
        

    def __init_subclass__(cls, **kwargs):  # @NoSelf
        """ register all subclassess to reconstruct them later """
        super().__init_subclass__(**kwargs)
        cls._subclasses[cls.__name__] = cls
        if hasattr(cls, 'name') and cls.name:
            if cls.name in cls._subclasses:
                logging.warn(f'Solver with name {cls.name} is already '
                             'registered')
            cls._subclasses[cls.name] = cls


    @classmethod
    def from_name(cls, name: str, pde: PDEBase, **kwargs) -> "SolverBase":
        r""" create solver class based on its name
        
        Solver classes are automatically registered when they inherit from
        :class:`SolverBase`. Note that this also requires that the respective
        python module containing the solver has been loaded before it is
        attempted to be used.
        
        Args:
            name (str):
                The name of the solver to construct
            pde (:class:`~pde.pdes.base.PDEBase`):
                The partial differential equation that should be solved
            \**kwargs:
                Additional arguments for the constructor of the solver
            
        Returns:
            An instance of a subclass of :class:`SolverBase`
        """
        try:
            # obtain the solver class associated with `name`
            solver_class = cls._subclasses[name]
        except KeyError:
            # solver was not registered
            solvers = (f"'{solver}'"
                       for solver in sorted(cls._subclasses.keys())
                       if not solver.endswith('Solver'))
            raise ValueError(f"Unknown solver method '{name}'. Registered "
                             f"solvers are {', '.join(solvers)}")

        return solver_class(pde, **kwargs)  # type: ignore
    
    
    @classproperty
    def registered_solvers(cls) -> List[str]:  # @NoSelf
        """ list of str: the names of the registered solvers """
        return list(sorted(cls._subclasses.keys()))
    

    def _make_pde_rhs(self, state: FieldBase,
                      backend: str = 'auto',
                      allow_stochastic: bool = False):
        """ obtain a function for evaluating the right hand side
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other
                information can be extracted
                The instance describing the pde that needs to be solved
            backend (str):
                Determines how the function is created. Accepted  values are
                'numpy` and 'numba'. Alternatively, 'auto' lets the code decide
                for the most optimal backend.
            allow_stochastic (bool):
                Flag indicating whether stochastic simulations should be
                supported.
                
        Raises:
            RuntimeError: when a stochastic partial differential equation is
            encountered but `allow_stochastic == False`.
        
        Returns:
            A function that is called with data given by a
            :class:`numpy.ndarray` and a time. The function returns the 
            deterministic evolution rate and (if applicable) a realization of
            the associated noise.
        """
        if self.pde.is_sde and not allow_stochastic:
            raise RuntimeError(f'The chosen stepper does not support '
                               'stochastic equations')
        
        if self.pde.is_sde:
            rhs = self.pde.make_sde_rhs(state, backend=backend)
        else:
            rhs = self.pde.make_pde_rhs(state, backend=backend)
            
        if hasattr(rhs, '_backend'):
            self.info['backend'] = rhs._backend  # type: ignore
        elif isinstance(rhs, nb.dispatcher.Dispatcher):
            self.info['backend'] = 'numba'
        else:           
            self.info['backend'] = 'undetermined'
        
        return rhs
            

    @abstractmethod
    def make_stepper(self, state, dt: float = None):
        pass
   




'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import pytest
import numpy as np

from .. import FileStorage
from ...grids import UnitGrid
from ...fields import ScalarField
from ...pdes import DiffusionPDE
from ...tools.misc import skipUnlessModule


        
@skipUnlessModule("h5py")
@pytest.mark.parametrize('compression', [True, False])
def test_storage_persistence(compression, tmp_path):
    """ test writing to persistent trackers """
    dim = 5
    grid = UnitGrid([dim])
    field = ScalarField(grid)

    path = tmp_path / f'test_storage_persistence_{compression}.hdf5'
    
    # write some data 
    for write_mode in ['append', 'truncate_once', 'truncate']:
        storage = FileStorage(path, info={'a': 1},
                              write_mode=write_mode,
                              compression=compression)
        
        # first batch
        storage.start_writing(field, info={'b': 2})
        storage.append(np.arange(dim), 0)
        storage.append(np.arange(dim, 2*dim))
        storage.end_writing()
        
        # read first batch
        np.testing.assert_array_equal(storage.times, np.arange(2))
        np.testing.assert_array_equal(np.ravel(storage.data), np.arange(10))
        assert {'a': 1, 'b': 2}.items() <= storage.info.items()
        
        # second batch
        storage.start_writing(field, info={'c': 3})
        storage.append(np.arange(2*dim, 3*dim), 2)
        storage.end_writing()
        
        storage.close()
        
        # read the data
        storage = FileStorage(path)
        if write_mode == 'truncate':
            np.testing.assert_array_equal(storage.times, np.array([2]))
            np.testing.assert_array_equal(np.ravel(storage.data),
                                          np.arange(10, 15))
            assert storage.shape == (1, 5)
            info = {'c': 3}
            assert info.items() <= storage.info.items()
        else:
            np.testing.assert_array_equal(storage.times, np.arange(0, 3))
            np.testing.assert_array_equal(np.ravel(storage.data),
                                          np.arange(0, 15))
            assert storage.shape == (3, 5)
            info = {'a': 1, 'b': 2, 'c': 3}
            assert info.items() <= storage.info.items()
    
    
     
@skipUnlessModule("h5py")
@pytest.mark.parametrize('compression', [True, False])
def test_simulation_persistence(compression, tmp_path):
    """ test whether a tracker can accurately store information about
    simulation """
    path = tmp_path / 'test_simulation_persistence.hdf5'
    storage = FileStorage(path, compression=compression)
     
    # write some simulation data
    pde = DiffusionPDE()
    grid = UnitGrid([16, 16])  # generate grid
    state = ScalarField.random_uniform(grid, 0.2, 0.3)
    pde.solve(state, t_range=0.11, dt=0.001,
              tracker=storage.tracker(interval=0.05))
    storage.close()        
     
    # read the data
    storage = FileStorage(path)
    np.testing.assert_almost_equal(storage.times, [0, 0.05, 0.1])
    data = np.array(storage.data)
    assert data.shape == (3,) + state.data.shape
    grid_res = storage.grid
    assert grid == grid_res
    grid_res = storage.grid
    assert grid == grid_res



@skipUnlessModule("h5py")
@pytest.mark.parametrize('compression', [True, False])
def test_storage_fixed_size(compression, tmp_path):
    """ test setting fixed size of FileStorage objects """
    c = ScalarField(UnitGrid([2]), data=1)

    for fixed in [True, False]:
        path = tmp_path / f'test_storage_fixed_size_{fixed}.hdf5'
        storage = FileStorage(path, max_length=1 if fixed else None,
                              compression=compression)
        assert len(storage) == 0
        
        storage.start_writing(c)
        assert len(storage) == 0
        storage.append(c.data, 0)
        assert len(storage) == 1
        
        if fixed:
            with pytest.raises((TypeError, ValueError)):
                storage.append(c.data, 1)
            assert len(storage) == 1
            np.testing.assert_allclose(storage.times, [0])
        else:
            storage.append(c.data, 1)
            assert len(storage) == 2
            np.testing.assert_allclose(storage.times, [0, 1])
            
            
            
@skipUnlessModule("h5py")
def test_appending(tmp_path):
    """ test the appending data """
    path = tmp_path / 'test_appending.hdf5'
    
    c = ScalarField(UnitGrid([2]), data=1)
    storage = FileStorage(path)
    storage.start_writing(c)
    assert len(storage) == 0
    storage.append(c.data, 0)
    assert storage._file_state == 'writing'
    assert len(storage) == 1
    storage.close()
                
    storage2 = FileStorage(path, write_mode='append')
    storage2.start_writing(c)
    storage2.append(c.data, 1)
    storage2.close()
    
    assert len(storage2) == 2
    

            
@skipUnlessModule("h5py")
def test_keep_opened(tmp_path):
    """ test the keep opened option """
    path = tmp_path / 'test_keep_opened.hdf5'
    
    c = ScalarField(UnitGrid([2]), data=1)
    storage = FileStorage(path, keep_opened=False)
    storage.start_writing(c)
    assert len(storage) == 0
    storage.append(c.data, 0)
    assert storage._file_state == 'closed'
    assert len(storage) == 1
    assert storage._file_state == 'reading'
    storage.append(c.data, 1)
    assert len(storage) == 2
                
    storage2 = FileStorage(path, write_mode='append')
    assert storage.times == storage2.times
    assert storage.data == storage2.data
    storage.close()  # close the old storage to enable writing here
    storage2.start_writing(c)
    storage2.append(c.data, 2)
    storage2.close()

    assert len(storage2) == 3
    np.testing.assert_allclose(storage2.times, np.arange(3))
    
                        


"""
Module defining classes for tracking results from simulations.

The trackers defined in this module are:

.. autosummary::
   :nosignatures:

   CallbackTracker
   ProgressTracker
   PrintTracker
   PlotTracker
   DataTracker
   SteadyStateTracker
   RuntimeTracker
   ConsistencyTracker
   MaterialConservationTracker

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from datetime import timedelta
import inspect
import sys
import os.path
import time
from typing import (Callable, Optional, Union, IO, List, Any,  # @UnusedImport
                    Dict, TYPE_CHECKING)

import numpy as np

from .base import TrackerBase, InfoDict, FinishedSimulation, Real
from .intervals import IntervalData, RealtimeIntervals
from ..fields.base import FieldBase
from ..fields import FieldCollection
from ..tools.parse_duration import parse_duration
from ..tools.misc import get_progress_bar_class
from ..tools.docstrings import fill_in_docstring



if TYPE_CHECKING:
    import pandas  # @UnusedImport



class CallbackTracker(TrackerBase):
    """ Tracker that calls a function periodically """
    
    @fill_in_docstring
    def __init__(self, func: Callable,
                 interval: IntervalData = 1):
        """ 
        Args:
            func: The function to call periodically. The function signature
                should be `(state)` or `(state, time)`, where `state` contains
                the current state as an instance of
                :class:`~pde.fields.FieldBase` and `time` is a
                float value indicating the current time. Note that only a view
                of the state is supplied, implying that a copy needs to be made
                if the data should be stored.
            interval:
                {ARG_TRACKER_INTERVAL}
        """
        super().__init__(interval=interval)
        self._callback = func
        self._num_args = len(inspect.signature(func).parameters)
        if not 0 < self._num_args < 3:
            raise ValueError('`func` must be a function accepting one or two '
                             f'arguments, not {self._num_args}') 
        
        
    def handle(self, field: FieldBase, t: float) -> None:
        """ handle data supplied to this tracker
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        if self._num_args == 1:
            self._callback(field)
        else:
            self._callback(field, t)



class ProgressTracker(TrackerBase):
    """ Tracker that shows the progress of the simulation """
            
    name = 'progress'


    @fill_in_docstring            
    def __init__(self, interval: IntervalData = None,
                 ndigits: int = 5, leave: bool = True):
        """
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
                The default value `None` updates the progress bar approximately
                every (real) second.
            ndigits (int): The number of digits after the decimal point that are
                shown maximally.
            leave (bool): Whether to leave the progress bar after the simulation
                has finished (default: True)
        """ 
        if interval is None:
            # print every second by default
            interval = RealtimeIntervals(duration=1)
        
        super().__init__(interval=interval)
        self.ndigits = ndigits
        self.leave = leave
        

    def initialize(self, field: FieldBase, info: InfoDict = None) -> float:
        """ initialize the tracker with information about the simulation
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation        
                
        Returns:
            float: The first time the tracker needs to handle data
        """
        result = super().initialize(field, info)
        
        # get solver information
        controller_info = {} if info is None else info.get('controller', {})
        
        # initialize the progress bar
        pb_cls = get_progress_bar_class()
        self.progress_bar = pb_cls(total=controller_info.get('t_end'),
                                   initial=controller_info.get('t_start', 0),
                                   leave=self.leave)
        self.progress_bar.set_description('Initializing')

        return result
            
            
    def handle(self, field: FieldBase, t: float) -> None:
        """ handle data supplied to this tracker
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        # show an update
        if self.progress_bar.total:
            t_new = min(t, self.progress_bar.total)
        else:
            t_new = t
        self.progress_bar.n = round(t_new, self.ndigits)
        self.progress_bar.set_description('')
                
        
    def finalize(self, info: InfoDict = None) -> None:
        """ finalize the tracker, supplying additional information

        Args:
            info (dict):
                Extra information from the simulation        
        """
        super().finalize(info)
        self.progress_bar.set_description('')

        # limit progress bar to 100%
        controller_info = {} if info is None else info.get('controller', {}) 
        t_final = controller_info.get('t_final', -np.inf)
        t_end = controller_info.get('t_end', -np.inf)
        if t_final >= t_end and self.progress_bar.total:
            self.progress_bar.n = self.progress_bar.total
            self.progress_bar.refresh()
        
        if (controller_info.get('successful', False) and self.leave and
                hasattr(self.progress_bar, 'sp')):
            # show progress bar in green if simulation was successful. We
            # need to overwrite the default behavior (and disable the
            # progress bar) since reaching steady state means the simulation
            # was successful even though it did not reach t_final
            try:
                self.progress_bar.sp(bar_style='success')
            except TypeError:
                self.progress_bar.close()
            else:
                self.disable = True
        else:
            self.progress_bar.close()
            
            
    def __del__(self):
        if hasattr(self, 'progress_bar') and not self.progress_bar.disable:
            self.progress_bar.close()



class PrintTracker(TrackerBase):
    """ Tracker that prints data to a stream (default: stdout) """
    
    name = 'print'
    
    
    @fill_in_docstring
    def __init__(self, interval: IntervalData = 1,
                 stream: IO[str] = sys.stdout):
        """
        
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
            stream:
                The stream used for printing
        """
        super().__init__(interval=interval)
        self.stream = stream
        
        
    def handle(self, field: FieldBase, t: float) -> None:
        """ handle data supplied to this tracker
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        data = f"c={field.data.mean():.3g}±{field.data.std():.3g}"
            
        self.stream.write(f"t={t:g}, {data}\n")
        self.stream.flush()



class PlotTracker(TrackerBase):
    """ Tracker that plots data on screen, to files, or writes a movie """
     
    name = 'plot'
    
    @fill_in_docstring
    def __init__(self, interval: IntervalData = 1,
                 title: str = '',
                 output_file: Optional[str] = None,
                 output_folder: Optional[str] = None,
                 movie_file: Optional[str] = None,
                 show: bool = True,
                 close_final: bool = False,
                 plot_args: Dict[str, Any] = None):
        """
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
            title (str):
                Text to show in the title. The current time point will be
                appended to this text, so include a space for optimal results.
            output_file (str, optional):
                Specifies a single image file, which is updated periodically, so
                that the progress can be monitored (e.g. on a compute cluster)
            output_folder (str, optional):
                Specifies a folder to which all images are written. The files
                will have names with increasing numbers.
            movie_file (str, optional):
                Specifies a filename to which a movie of all the frames is
                written after the simulation.
            show (bool, optional):
                Determines whether the plot is shown while the simulation is
                running. If `False`, the files are created in the background.
                This option can slow down a simulation severely.
            plot_args (dict):
                Extra arguments supplied to the plot call
        """
        super().__init__(interval=interval)
        self.title = title
        self.output_file = output_file
        self.output_folder = output_folder
        self.show = show
        self.close_final = close_final
        
        self.plot_args = {} if plot_args is None else plot_args
        self.plot_args['show'] = False  # this is handled by the context
        
        if movie_file is not None or output_folder is not None:
            from ..visualization.movies import Movie
            movie = Movie(filename=movie_file, image_folder=output_folder)
            self.movie: Optional[Movie] = movie
            self.movie._start()  # initialize movie
        else:
            self.movie = None
         
     
    def initialize(self, state: FieldBase, info: InfoDict = None) -> float:
        """ initialize the tracker with information about the simulation
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation
                
        Returns:
            float: The first time the tracker needs to handle data
        """
        # initialize the plotting context
        from ..tools.plotting import get_plotting_context
        title = self.title + 'Initializing...'
        self._context = get_plotting_context(title=title, show=self.show)

        # do the actual plotting
        with self._context:
            self._plot_reference = state.plot(**self.plot_args)

        if self._context.supports_update:
            # the context supports reusing figures
            if hasattr(state.plot, 'update_method'):
                # the plotting method supports updating the plot
                if state.plot.update_method is None:  # type: ignore
                    if state.plot.mpl_class == 'axes':  # type: ignore
                        self._update_method = 'update_ax'
                    elif state.plot.mpl_class == 'figure':  # type: ignore
                        self._update_method = 'update_fig'
                    else:
                        mpl_class = state.plot.mpl_class  # type: ignore
                        raise RuntimeError('Unknown mpl_class on plot method: '
                                           f'{mpl_class}')
                else: 
                    self._update_method = 'update_data'
            else:
                raise RuntimeError('PlotTracker does not  work since the state '
                                   f'of type {state.__class__.__name__} does '
                                   'not use the plot protocol of '
                                   '`pde.tools.plotting`.')
        else:
            self._update_method = 'replot'
            
        self._logger.info(f'Update method: "{self._update_method}"')
                
        return super().initialize(state, info=info)
        
         
    def handle(self, state: FieldBase, t: float) -> None:
        """ handle data supplied to this tracker
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        self._context.title = f'{self.title}Time: {t:g}'
        
        # update the plot in the correct plotting context
        with self._context:
            if self._update_method == 'update_data':
                # the state supports updating the plot data
                update_func = getattr(state,
                                      state.plot.update_method)  # type: ignore
                update_func(self._plot_reference)
                
            elif self._update_method == 'update_fig':
                fig = self._context.fig
                fig.clf()  # type: ignore
                state.plot(fig=fig, **self.plot_args)
                
            elif self._update_method == 'update_ax':
                fig = self._context.fig
                fig.clf()  # type: ignore
                ax = fig.add_subplot(1, 1, 1)  # type: ignore
                state.plot(ax=ax, **self.plot_args)
                
            elif self._update_method == 'replot':
                state.plot(**self.plot_args)
                
            else:
                raise RuntimeError('Unknown update method '
                                   f'`{self._update_method}`')
                
        if self.output_file and self._context.fig is not None:
            self._context.fig.savefig(self.output_file)
        if self.movie:
            self.movie.add_figure(self._context.fig)
          

    def finalize(self, info: InfoDict = None) -> None:
        """ finalize the tracker, supplying additional information

        Args:
            info (dict):
                Extra information from the simulation        
        """
        super().finalize(info)
        if self.movie:
            if self.movie.filename:
                # write out movie file if requested
                self._logger.info(f'Writing movie to {self.movie.filename}...')
                self.movie.save()
            # finalize movie (e.g. delete temporary files)
            self.movie._end()
            
        if not self.show or self.close_final:
            self._context.close()
    
              
    
class DataTracker(CallbackTracker):
    """ Tracker that stores custom data obtained by calling a function
    
    Attributes:
        times (list):
            The time points at which the data is stored
        data (list):
            The actually stored data, which is a list of the objects returned by
            the callback function. 
    """
    
    @fill_in_docstring
    def __init__(self, func: Callable,
                 interval: IntervalData = 1,
                 filename: str = None):
        """ 
        Args:
            func:
                The function to call periodically. The function signature
                should be `(state)` or `(state, time)`, where `state` contains
                the current state as an instance of
                :class:`~pde.fields.FieldBase` and `time` is a
                float value indicating the current time. Note that only a view
                of the state is supplied, implying that a copy needs to be made
                if the data should be stored.
                Typical return values of the function are either a single
                number, a numpy array, a list of number, or a dictionary to
                return multiple numbers with assigned labels.
            interval:
                {ARG_TRACKER_INTERVAL}
            filename (str):
                A path to a file to which the data is written at the end of the
                tracking. The data format will be determined by the extension
                of the filename. '.pickle' indicates a python pickle file
                storing a tuple `(self.times, self.data)`, whereas any other
                data format requires :mod:`pandas`.  
        """
        super().__init__(func=func, interval=interval)
        self.filename = filename
        self.times: List[float] = []
        self.data: List[Any] = []
        
        
    def handle(self, field: FieldBase, t: float) -> None:
        """ handle data supplied to this tracker
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        self.times.append(t)
        if self._num_args == 1:
            self.data.append(self._callback(field))
        else:
            self.data.append(self._callback(field, t))
        
        
    def finalize(self, info: InfoDict = None) -> None:
        """ finalize the tracker, supplying additional information

        Args:
            info (dict):
                Extra information from the simulation        
        """
        super().finalize(info)
        if self.filename:
            self.to_file(self.filename)
        
        
    @property
    def dataframe(self) -> "pandas.DataFrame":
        """ :class:`pandas.DataFrame`: the data in a dataframe
        
        If `func` returns a dictionary, the keys are used as column names.
        Otherwise, the returned data is enumerated starting with '0'. In any
        case the time point at which the data was recorded is stored in the
        column 'time'.
        """
        import pandas as pd
        df = pd.DataFrame(self.data)
        # insert the times and use them as an index
        df.insert(0, 'time', self.times)
        return df
    
    
    def to_file(self, filename: str, **kwargs):
        r""" store data in a file
        
        The extension of the filename determines what format is being used. For
        instance, '.pickle' indicates a python pickle file storing a tuple
        `(self.times, self.data)`, whereas any other data format requires
        :mod:`pandas`. Supported formats include 'csv', 'json'.  
                
        Args:
            filename (str):
                Path where the data is stored
            \**kwargs:
                Additional parameters may be supported for some formats 
        """
        extension = os.path.splitext(filename)[1].lower()
        if extension == '.pickle':
            # default 
            import pickle
            with open(filename, "wb") as fp:
                pickle.dump((self.times, self.data), fp, **kwargs)
            
        elif extension == '.csv':
            self.dataframe.to_csv(filename, **kwargs)
        elif extension == '.json':
            self.dataframe.to_json(filename, **kwargs)
        elif extension in {'.xls', '.xlsx'}:
            self.dataframe.to_excel(filename, **kwargs)
        else:
            raise ValueError(f'Unsupported file extension `{extension}`')
            
            
            
class SteadyStateTracker(TrackerBase):
    """ Tracker that interrupts the simulation once steady state is reached
    
    Steady state is obtained when the state does not change anymore. This is the
    case when the derivative is close to zero.
    """

    name = 'steady_state'


    @fill_in_docstring
    def __init__(self, interval: IntervalData = None,
                 atol: float = 1e-8,
                 rtol: float = 1e-5):
        """
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
                The default value `None` checks for the steady state
                approximately every (real) second.
            atol (float): Absolute tolerance that must be reached to abort the
                simulation
            rtol (float): Relative tolerance that must be reached to abort the
                simulation
        """ 
        if interval is None:
            interval = RealtimeIntervals(duration=1)
        super().__init__(interval=interval)
        self.atol = atol 
        self.rtol = rtol
        self._last_data = None
        
        
    def handle(self, field: FieldBase, t: float) -> None:
        """ handle data supplied to this tracker
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        if self._last_data is not None:
            # scale with dt to make test independent of dt
            atol = self.atol * self.interval.dt
            rtol = self.rtol * self.interval.dt
            if np.allclose(self._last_data, field.data,
                           rtol=rtol, atol=atol, equal_nan=True):
                raise FinishedSimulation('Reached stationary state')
            
        self._last_data = field.data.copy()  # store data from last timestep
            


class RuntimeTracker(TrackerBase):
    """ Tracker that interrupts the simulation once a duration has passed """


    @fill_in_docstring
    def __init__(self, max_runtime: Union[Real, str],
                 interval: IntervalData = 1):  
        """
        Args:
            max_runtime (float or str):
                The maximal runtime of the simulation. If the runtime is
                exceeded, the simulation is interrupted. Values can be either
                given as a number (interpreted as seconds) or as a string, which
                is then parsed using the function
                :func:`~pde.tools.parse_duration.parse_duration`.
            interval:
                {ARG_TRACKER_INTERVAL}
        """
        super().__init__(interval=interval)
        
        try:
            self.max_runtime = float(max_runtime)
        except ValueError:
            td = parse_duration(str(max_runtime))
            self.max_runtime = td.total_seconds()


    def initialize(self, field: FieldBase, info: InfoDict = None) -> float:
        """ 
        Args:
            field (:class:`~pde.fields.FieldBase`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation        
                
        Returns:
            float: The first time the tracker needs to handle data
        """
        self.max_time = time.time() + self.max_runtime
        return super().initialize(field, info)
        
        
    def handle(self, field: FieldBase, t: float) -> None:
        """ handle data supplied to this tracker
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        if time.time() > self.max_time:
            dt = timedelta(seconds=self.max_runtime)
            raise FinishedSimulation(f'Reached maximal runtime of {str(dt)}')

            
            
class ConsistencyTracker(TrackerBase):
    """ Tracker that interrupts the simulation when the state is not finite """ 

    name = 'consistency'
        
    
    @fill_in_docstring
    def __init__(self, interval: IntervalData = None):
        """
        Args:
            interval: 
                {ARG_TRACKER_INTERVAL}
                The default value `None` checks for consistency approximately
                every (real) second.
        """ 
        if interval is None:
            interval = RealtimeIntervals(duration=1)
        super().__init__(interval=interval)
        
        
    def handle(self, field: FieldBase, t: float) -> None:
        """ handle data supplied to this tracker
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        if not np.all(np.isfinite(field.data)):
            raise StopIteration('Field was not finite')
            
        self._last = field.data.copy()  # store data from last timestep
            


class MaterialConservationTracker(TrackerBase):
    """ Ensure that the amount of material is conserved """

    name = 'material_conservation'


    @fill_in_docstring
    def __init__(self, interval: IntervalData = 1,
                 atol: float = 1e-4,
                 rtol: float = 1e-4):
        """
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
            atol (float):
                Absolute tolerance for amount deviations
            rtol (float):
                Relative tolerance for amount deviations
        """
        super().__init__(interval=interval)
        self.atol = atol 
        self.rtol = rtol
        
        
    def initialize(self, field: FieldBase, info: InfoDict = None) -> float:
        """ 
        Args:
            field (:class:`~pde.fields.base.FieldBase`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation        
                
        Returns:
            float: The first time the tracker needs to handle data
        """
        if isinstance(field, FieldCollection):
            self._reference = np.array([f.magnitude for f in field])
        else:
            self._reference = field.magnitude  # type: ignore
            
        return super().initialize(field, info)
        
        
    def handle(self, field: FieldBase, t: float) -> None:
        """ handle data supplied to this tracker
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        if isinstance(field, FieldCollection):
            mags = np.array([f.magnitude for f in field])
        else:
            mags = field.magnitude  # type: ignore
            
        c = np.isclose(mags, self._reference, rtol=self.rtol, atol=self.atol)
        if not np.all(c):
            if isinstance(field, FieldCollection):
                msg = f'Material of field {np.flatnonzero(~c)} is not conserved'
            else:
                msg = f'Material is not conserved'
            raise StopIteration(msg)
            
            
            
__all__ = ['CallbackTracker', 'ProgressTracker', 'PrintTracker', 'PlotTracker',
           'DataTracker', 'SteadyStateTracker', 'RuntimeTracker',
           'ConsistencyTracker', 'MaterialConservationTracker']



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import itertools

import numpy as np
import pytest

from ... import grids
from ..base import discretize_interval, GridBase
from ...tools.misc import skipUnlessModule



def iter_grids():
    """ generate some test grids """
    yield grids.UnitGrid([2, 2], periodic=[True, False])
    yield grids.CartesianGrid([[0, 1]], [2], periodic=[False])
    yield grids.CylindricalGrid(2, (0, 2), (2, 2), periodic_z=True)
    yield grids.SphericalGrid(2, 2)
    yield grids.PolarGrid(2, 2)



def test_discretize():
    """ test the discretize function """
    x_min = np.random.uniform(0, 1)
    x_max = np.random.uniform(2, 3)
    num = np.random.randint(5, 8)
    x, dx = discretize_interval(x_min, x_max, num)
    assert dx == pytest.approx((x_max - x_min) / num)
    x_expect = np.linspace(x_min + dx / 2, x_max - dx / 2, num)
    np.testing.assert_allclose(x, x_expect)



def test_serialization():
    """ test whether grid can be serialized """
    def iter_grids():
        """ helper function iterating over different grids """
        for periodic in [True, False]:
            yield grids.UnitGrid([3, 4], periodic=periodic)
            yield grids.CartesianGrid([[0, 1], [-2, 3]], [4, 5],
                                      periodic=periodic)
            yield grids.CylindricalGrid(3, [-1, 2], [5, 7],
                                        periodic_z=periodic)
        yield grids.SphericalGrid(4, 6)
        yield grids.PolarGrid(4, 5)
            
    for grid in iter_grids():
        g = GridBase.from_state(grid.state_serialized)
        assert grid == g
        assert grid._cache_hash() == g._cache_hash()
            
        
        
def test_iter_mirror_points():
    """ test iterating mirror points in grids """
    grid_cart = grids.UnitGrid([2, 2], periodic=[True, False])
    grid_cyl = grids.CylindricalGrid(2, (0, 2), (2, 2), periodic_z=False)
    grid_sph = grids.SphericalGrid(2, 2)
    assert (grid_cart._cache_hash() != grid_cyl._cache_hash() != 
            grid_sph._cache_hash())
    
    for with_, only_periodic in itertools.product([False, True], repeat=2):
        num_expect = 2 if only_periodic else 8
        num_expect += 1 if with_ else 0
        ps = grid_cart.iter_mirror_points([1, 1], with_, only_periodic)
        assert len(list(ps)) == num_expect
        
        num_expect = 0 if only_periodic else 2
        num_expect += 1 if with_ else 0
        ps = grid_cyl.iter_mirror_points([0, 0, 1], with_,
                                         only_periodic)
        assert len(list(ps)) == num_expect
        
        num_expect = 1 if with_ else 0
        ps = grid_sph.iter_mirror_points([0, 0, 0], with_,
                                         only_periodic)
        assert len(list(ps)) == num_expect
        
        
        
def test_cell_to_point_conversion():
    """ test the conversion between cells and points """
    for grid in iter_grids():
        c = grid.point_to_cell(grid.get_random_point())
        c2 = grid.point_to_cell(grid.cell_to_point(c))
        np.testing.assert_almost_equal(c, c2)
        
        p_emtpy = np.zeros((0, grid.num_axes))
        assert grid.point_to_cell(p_emtpy).size == 0
        assert grid.cell_to_point(p_emtpy).size == 0
        
        

@skipUnlessModule('matplotlib')
def test_grid_plotting():
    """ test plotting of grids """
    grids.UnitGrid([4]).plot()
    grids.UnitGrid([4, 4]).plot()
    
    with pytest.raises(NotImplementedError):
        grids.UnitGrid([4, 4, 4]).plot()

    grids.PolarGrid(4, 8).plot()
    grids.PolarGrid((2, 4), 8).plot()
    
    
    
def test_operators():
    """ test operator mechanism """
    def make_op(state):
        return lambda state: state
    for grid in iter_grids():
        assert "laplace" in grid.operators
        with pytest.raises(ValueError):
            grid.get_operator('not_existent', 'natural')
        grid.register_operator('noop', make_op)
        assert "noop" in grid.operators
        del grid._operators['noop']  # reset original state
        


#!/usr/bin/env python3
'''
Code for creating performance plots

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import os
os.environ['NUMBA_NUM_THREADS'] = '1'  # check single thread performance

import functools
import timeit

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from pde import UnitGrid
from pde.tools.misc import display_progress, estimate_computation_speed

try:
    import cv2
except ImportError:
    opencv_laplace = None
else:
    opencv_laplace = functools.partial(cv2.Laplacian, ddepth=cv2.CV_64F,
                                       borderType=cv2.BORDER_REFLECT)



def time_function(func, arg, repeat=3):
    """ estimates the computation speed of a function
    
    Args:
        func (callable): The function to test
        arg: The single argument on which the function will be estimate
        repeat (int): How often the function is tested
        
    Returns:
        float: Estimated duration of calling the function a single time
    """
    number = int(estimate_computation_speed(func, arg))
    func = functools.partial(func, arg)
    return min(timeit.repeat(func, number=number, repeat=repeat)) / number
    


def get_performance_data(periodic=False):
    """ obtain the data used in the performance plot
    
    Args:
        periodic (bool): The boundary conditions of the underlying grid
        
    Returns:
        dict: The durations of calculating the Laplacian on different grids
        using different methods
    """
    sizes = 2 ** np.arange(3, 13)

    statistics = {}
    for size in display_progress(sizes):
        data = {}
        grid = UnitGrid([size] * 2, periodic=periodic)
        test_data = np.random.randn(*grid.shape)
        
        for method in ['numba', 'scipy']:
            op = grid.get_operator('laplace', bc='natural', method=method)
            data[method] = time_function(op, test_data)
            
        if opencv_laplace:
            data['opencv'] = time_function(opencv_laplace, test_data)
            
        statistics[int(size)] = data

    return statistics



def plot_performance(performance_data, title=None):
    """ plot the performance data
    
    Args:
        performance_data: The data obtained from calling
            :func:`get_performance_data`.
        title (str): The title of the plot
    """
    plt.figure(figsize=[4, 3])
    
    METHOD_LABELS = {'numba': 'py-pde'}
    
    sizes = np.array(sorted(performance_data.keys()))
    grid_sizes = sizes ** 2
    methods = sorted(performance_data[sizes[0]].keys(), reverse=True)
    
    for method in methods:
        data = np.array([performance_data[size][method] for size in sizes])
        plt.loglog(grid_sizes, data, '.-',
                   label=METHOD_LABELS.get(method, method))
        
    plt.xlim(grid_sizes[0], grid_sizes[-1])
    plt.xlabel('Number of grid points')
    plt.ylabel('Runtime [ms]')
    plt.legend(loc='best')
    
    # fix ticks of y-axis
    locmaj = mpl.ticker.LogLocator(base=10, numticks=12) 
    plt.gca().xaxis.set_major_locator(locmaj)
    
    if title:
        plt.title(title)
        
    plt.tight_layout()



def main():
    """ run main scripts """
    data = get_performance_data(periodic=False)
    plot_performance(data, title="2D Laplacian (reflecting BCs)")
    plt.savefig('performance_noflux.pdf', transparent=True)
    plt.savefig('performance_noflux.png', transparent=True, dpi=200)
    plt.close()
    
    data = get_performance_data(periodic=True)
    plot_performance(data, title="2D Laplacian (periodic BCs)")
    plt.savefig('performance_periodic.pdf', transparent=True)
    plt.savefig('performance_periodic.png', transparent=True, dpi=200)
    plt.close()
    


if __name__ == "__main__":
    main()
    


r"""
This module implements differential operators on cylindrical grids 

.. autosummary::
   :nosignatures:

   make_laplace
   make_gradient
   make_divergence
   make_vector_gradient
   make_vector_laplace
   make_tensor_divergence
   
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from typing import Callable

from .common import PARALLELIZATION_THRESHOLD_2D
from .. import CylindricalGrid
from ..boundaries import Boundaries
from ...tools.numba import nb, jit_allocate_out
from ...tools.docstrings import fill_in_docstring

        
        
@fill_in_docstring
def make_laplace(bcs: Boundaries) -> Callable:
    """ make a discretized laplace operator for a cylindrical grid
    
    {DESCR_CYLINDRICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, CylindricalGrid)
    bcs.check_value_rank(0)
    boundary_r, boundary_z = bcs

    # calculate preliminary quantities
    dim_r, dim_z = bcs.grid.shape
    dr_2, dz_2 = 1 / bcs.grid.discretization**2
    
    value_outer = boundary_r.high.make_virtual_point_evaluator()
    region_z = boundary_z.make_region_evaluator()
    
    # use processing for large enough arrays 
    parallel = (dim_r * dim_z >= PARALLELIZATION_THRESHOLD_2D**2)

    @jit_allocate_out(parallel=parallel, out_shape=(dim_r, dim_z))
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        for j in nb.prange(0, dim_z):  # iterate axial points
            # inner radial boundary condition
            i = 0
            arr_z_l, arr_c, arr_z_h = region_z(arr, (i, j))
            out[i, j] = (
                2 * (arr[i + 1, j] - arr_c) * dr_2 +
                (arr_z_l - 2 * arr_c + arr_z_h) * dz_2
            )
            
            if dim_r == 1:
                continue  # deal with singular radial dimension
            
            for i in range(1, dim_r - 1):  # iterate radial points
                arr_z_l, arr_c, arr_z_h = region_z(arr, (i, j))
                arr_r_l, arr_r_h = arr[i - 1, j], arr[i + 1, j]
                out[i, j] = (
                    (arr_r_h - 2 * arr_c + arr_r_l) * dr_2 +
                    (arr_r_h - arr_r_l) / (2 * i + 1) * dr_2 +
                    (arr_z_l - 2 * arr_c + arr_z_h) * dz_2
                )
                
            # outer radial boundary condition
            i = dim_r - 1
            arr_z_l, arr_c, arr_z_h = region_z(arr, (i, j))
            arr_r_l, arr_r_h = arr[i - 1, j], value_outer(arr, (i, j))
            out[i, j] = (
                (arr_r_h - 2 * arr_c + arr_r_l) * dr_2 +
                (arr_r_h - arr_r_l) / (2 * i + 1) * dr_2 +
                (arr_z_l - 2 * arr_c + arr_z_h) * dz_2
            )
        return out
        
    return laplace  # type: ignore



@fill_in_docstring
def make_gradient(bcs: Boundaries) -> Callable:
    """ make a discretized gradient operator for a cylindrical grid
    
    {DESCR_CYLINDRICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, CylindricalGrid)
    bcs.check_value_rank(0)
    boundary_r, boundary_z = bcs

    # calculate preliminary quantities
    dim_r, dim_z = bcs.grid.shape
    scale_r, scale_z = 1 / (2 * bcs.grid.discretization)
    
    value_outer = boundary_r.high.make_virtual_point_evaluator()
    region_z = boundary_z.make_region_evaluator()

    # use processing for large enough arrays 
    parallel = (dim_r * dim_z >= PARALLELIZATION_THRESHOLD_2D**2)

    @jit_allocate_out(parallel=parallel, out_shape=(3, dim_r, dim_z))
    def gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        for j in nb.prange(0, dim_z):  # iterate axial points
            # inner radial boundary condition
            i = 0
            arr_z_l, _, arr_z_h = region_z(arr, (i, j))
            out[0, i, j] = (arr[1, i] - arr[0, i]) * scale_r
            out[1, i, j] = (arr_z_h - arr_z_l) * scale_z
            out[2, i, j] = 0  # no phi dependence by definition
            
            for i in range(1, dim_r - 1):  # iterate radial points
                arr_z_l, _, arr_z_h = region_z(arr, (i, j))
                out[0, i, j] = (arr[i + 1, j] - arr[i - 1, j]) * scale_r
                out[1, i, j] = (arr_z_h - arr_z_l) * scale_z
                out[2, i, j] = 0  # no phi dependence by definition
                
            # outer radial boundary condition
            i = dim_r - 1
            arr_z_l, _, arr_z_h = region_z(arr, (i, j))
            arr_r_h = value_outer(arr, (i, j))
            out[0, i, j] = (arr_r_h - arr[i - 1, j]) * scale_r
            out[1, i, j] = (arr_z_h - arr_z_l) * scale_z
            out[2, i, j] = 0  # no phi dependence by definition
            
        return out
        
    return gradient  # type: ignore



@fill_in_docstring
def make_divergence(bcs: Boundaries) -> Callable:
    """ make a discretized divergence operator for a cylindrical grid
    
    {DESCR_CYLINDRICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, CylindricalGrid)
    bcs.check_value_rank(0)
    boundary_r, boundary_z = bcs

    # calculate preliminary quantities
    dim_r, dim_z = bcs.grid.shape
    dr = bcs.grid.discretization[0]
    scale_r, scale_z = 1 / (2 * bcs.grid.discretization)
    
    value_outer = boundary_r.high.make_virtual_point_evaluator()
    region_z = boundary_z.make_region_evaluator()

    # use processing for large enough arrays 
    parallel = (dim_r * dim_z >= PARALLELIZATION_THRESHOLD_2D**2)

    @jit_allocate_out(parallel=parallel, out_shape=(dim_r, dim_z))
    def divergence(arr, out=None):
        """ apply divergence operator to array `arr` """            
        for j in nb.prange(0, dim_z):  # iterate axial points
            # inner radial boundary condition
            i = 0
            arr_z_l, _, arr_z_h = region_z(arr[1], (i, j))
            d_r = (arr[0, 1, j] + 3 * arr[0, 0, j]) * scale_r
            d_z = (arr_z_h - arr_z_l) * scale_z
            out[i, j] = d_r + d_z
            
            for i in range(1, dim_r - 1):  # iterate radial points
                arr_z_l, _, arr_z_h = region_z(arr[1], (i, j))
                d_r = (arr[0, i + 1, j] - arr[0, i - 1, j]) * scale_r + \
                      (arr[0, i, j] / ((i + 0.5) * dr))
                d_z = (arr_z_h - arr_z_l) * scale_z
                out[i, j] = d_r + d_z
                
            # outer radial boundary condition
            i = dim_r - 1
            arr_z_l, _, arr_z_h = region_z(arr[1], (i, j))
            arr_r_h = value_outer(arr[0], (i, j))
            d_r = (arr_r_h - arr[0, i - 1, j]) * scale_r + \
                  (arr[0, i, j] / ((i + 0.5) * dr))
            d_z = (arr_z_h - arr_z_l) * scale_z
            out[i, j] = d_z + d_r
            
        return out
        
    return divergence  # type: ignore



@fill_in_docstring
def make_vector_gradient(bcs: Boundaries) -> Callable:
    """ make a discretized vector gradient operator for a cylindrical grid
    
    {DESCR_CYLINDRICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, CylindricalGrid)
    bcs.check_value_rank(1)

    # calculate preliminary quantities
    gradient_r = make_gradient(bcs.extract_component(0))
    gradient_z = make_gradient(bcs.extract_component(1))
    gradient_phi = make_gradient(bcs.extract_component(2))
        
    @jit_allocate_out(out_shape=(3, 3) + bcs.grid.shape)
    def vector_gradient(arr, out=None):
        """ apply gradient operator to array `arr` """
        gradient_r(arr[0], out=out[:, 0])
        gradient_z(arr[1], out=out[:, 1])
        gradient_phi(arr[2], out=out[:, 2])
        return out    
        
    return vector_gradient  # type: ignore



@fill_in_docstring
def make_vector_laplace(bcs: Boundaries) -> Callable:
    """ make a discretized vector laplace operator for a cylindrical grid
    
    {DESCR_CYLINDRICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, CylindricalGrid)
    bcs.check_value_rank(1)

    laplace_r = make_laplace(bcs.extract_component(0))
    laplace_z = make_laplace(bcs.extract_component(1))
    laplace_phi = make_laplace(bcs.extract_component(2))
        
    @jit_allocate_out(out_shape=(3,) + bcs.grid.shape)
    def vector_laplace(arr, out=None):
        """ apply gradient operator to array `arr` """
        laplace_r(arr[0], out=out[0])
        laplace_z(arr[1], out=out[1])
        laplace_phi(arr[2], out=out[2])
        return out    
        
    return vector_laplace  # type: ignore



@fill_in_docstring
def make_tensor_divergence(bcs: Boundaries) -> Callable:
    """ make a discretized tensor divergence operator for a cylindrical grid
    
    {DESCR_CYLINDRICAL_GRID}

    Args:
        bcs (:class:`~pde.grids.boundaries.axes.Boundaries`):
            {ARG_BOUNDARIES_INSTANCE}
        
    Returns:
        A function that can be applied to an array of values
    """
    assert isinstance(bcs.grid, CylindricalGrid)
    bcs.check_value_rank(1)

    divergence_r = make_divergence(bcs.extract_component(0))
    divergence_z = make_divergence(bcs.extract_component(1))
    divergence_phi = make_divergence(bcs.extract_component(2))
        
    @jit_allocate_out(out_shape=(3,) + bcs.grid.shape)
    def tensor_divergence(arr, out=None):
        """ apply gradient operator to array `arr` """
        divergence_r(arr[0], out=out[0])
        divergence_z(arr[1], out=out[1])
        divergence_phi(arr[2], out=out[2])
        return out
        
    return tensor_divergence  # type: ignore


# register all operators with the grid class
CylindricalGrid.register_operator('laplace', make_laplace,
                                  rank_in=0, rank_out=0)
CylindricalGrid.register_operator('gradient', make_gradient,
                                  rank_in=0, rank_out=1)
CylindricalGrid.register_operator('divergence', make_divergence,
                                  rank_in=1, rank_out=0)
CylindricalGrid.register_operator('vector_gradient', make_vector_gradient,
                                  rank_in=1, rank_out=2)
CylindricalGrid.register_operator('vector_laplace', make_vector_laplace,
                                  rank_in=1, rank_out=1)
CylindricalGrid.register_operator('tensor_divergence', make_tensor_divergence,
                                  rank_in=2, rank_out=1)




"""
A simple diffusion equation

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import Callable

import numpy as np


from .base import PDEBase
from ..fields import ScalarField
from ..grids.boundaries.axes import BoundariesData
from ..tools.numba import nb, jit
from ..tools.docstrings import fill_in_docstring


        
class DiffusionPDE(PDEBase):
    r""" A simple diffusion equation
    
    The mathematical definition is

    .. math::
        \partial_t c = D \nabla^2 c
        
    where :math:`c` is a scalar field that is distributed with diffusivity
    :math:`D`.
    """

    explicit_time_dependence = False
    
    
    @fill_in_docstring
    def __init__(self, diffusivity: float = 1,
                 noise: float = 0,
                 bc: BoundariesData = 'natural'):
        """ 
        Args:
            diffusivity (float):
                The diffusivity of the described species
            noise (float):
                Strength of the (additive) noise term
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES} 
        """
        super().__init__(noise=noise)
        
        self.diffusivity = diffusivity
        self.bc = bc
            
            
    def evolution_rate(self, state: ScalarField,  # type: ignore
                       t: float = 0) -> ScalarField:
        """ evaluate the right hand side of the PDE
        
        Args:
            state (:class:`~pde.fields.ScalarField`):
                The scalar field describing the concentration distribution
            t (float): The current time point
            
        Returns:
            :class:`~pde.fields.ScalarField`:
            Scalar field describing the evolution rate of the PDE 
        """
        assert isinstance(state, ScalarField)
        laplace = state.laplace(bc=self.bc, label='evolution rate')
        return self.diffusivity * laplace  # type: ignore
    
    
    def _make_pde_rhs_numba(self, state: ScalarField  # type: ignore
                            ) -> Callable:
        """ create a compiled function evaluating the right hand side of the PDE
        
        Args:
            state (:class:`~pde.fields.ScalarField`):
                An example for the state defining the grid and data types
                
        Returns:
            A function with signature `(state_data, t)`, which can be called
            with an instance of :class:`numpy.ndarray` of the state data and
            the time to obtained an instance of :class:`numpy.ndarray` giving
            the evolution rate.  
        """
        shape = state.grid.shape
        arr_type = nb.typeof(np.empty(shape, dtype=np.double))
        signature = arr_type(arr_type, nb.double)
        
        diffusivity_value = self.diffusivity
        laplace = state.grid.get_operator('laplace', bc=self.bc)

        @jit(signature)
        def pde_rhs(state_data: np.ndarray, t: float):
            """ compiled helper function evaluating right hand side """ 
            return diffusivity_value * laplace(state_data) 
            
        return pde_rhs  # type: ignore
    
    
    



'''
Common functions that are used by many operators 

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from typing import Callable
import logging
import warnings

import numpy as np
from scipy import sparse



# Package-wide constant defining when to use parallel numba 
PARALLELIZATION_THRESHOLD_2D = 256
""" int: threshold for determining when parallel code is created for
differential operators. The value gives the minimal number of support points in
each direction for a 2-dimensional grid """
PARALLELIZATION_THRESHOLD_3D = 64
""" int: threshold for determining when parallel code is created for
differential operators. The value gives the minimal number of support points in
each direction for a 3-dimensional grid """



logger = logging.getLogger(__name__) 



def make_laplace_from_matrix(matrix, vector) -> Callable:
    """ make a Laplace operator using matrix vector products

    Args:
        matrix:
            The (sparse) matrix representing the laplace operator on the given
            grid.
        vector:
            The constant part representing the boundary conditions of the
            Laplace operator.
        
    Returns:
        A function that can be applied to an array of values to obtain the
        solution to Poisson's equation where the array is used as the right hand
        side
    """
    mat = matrix.tocsc()
    vec = vector.toarray()[:, 0]
    
    def laplace(arr: np.ndarray, out: np.ndarray = None) -> np.ndarray:
        """ apply the laplace operator to `arr` """
        result = mat.dot(arr.flat) + vec
        if out is None:
            return result.reshape(arr.shape)
        else:
            out.flat[:] = result
            return out    
    
    return laplace



def make_general_poisson_solver(matrix, vector, method: str = 'auto') \
        -> Callable:
    """ make an operator that solves Poisson's problem

    Args:
        matrix:
            The (sparse) matrix representing the laplace operator on the given
            grid.
        vector:
            The constant part representing the boundary conditions of the
            Laplace operator.
        method (str):
            The chosen method for implementing the operator
        
    Returns:
        A function that can be applied to an array of values to obtain the
        solution to Poisson's equation where the array is used as the right hand
        side
    """
    if method not in {'auto', 'scipy'}:
        raise ValueError(f'Method {method} is not available')
    
    # prepare the matrix representing the operator
    mat = matrix.tocsc()
    vec = vector.toarray()[:, 0]
    
    def solve_poisson(arr: np.ndarray, out: np.ndarray = None) -> np.ndarray:
        """ solves Poisson's equation using sparse linear algebra """
        # prepare the right hand side vector
        rhs = arr.flat - vec
        
        # solve the linear problem using a sparse solver
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("error")  # enable warning catching
                result = sparse.linalg.spsolve(mat, rhs)
                
        except sparse.linalg.dsolve.linsolve.MatrixRankWarning:
            # this can happen for singular laplace matrix, e.g. when pure
            # Neumann conditions are considered. In this case, a solution is
            # obtained using least squares
            logger.warning('Poisson problem seems to be under-determined and '
                           'could not be solved using sparse.linalg.spsolve')
            use_leastsquares = True
            
        else:
            # test whether the solution is good enough
            if np.allclose(mat.dot(result), rhs, rtol=1e-5, atol=1e-5):
                logger.info('Solved Poisson problem with sparse.linalg.spsolve')
                use_leastsquares = False
            else:
                logger.warning('Poisson problem was not solved using '
                               'sparse.linalg.spsolve')
                use_leastsquares = True
                
        if use_leastsquares:
            # use least squares to solve an underdetermined problem
            result = sparse.linalg.lsmr(mat, rhs)[0]
            if not np.allclose(mat.dot(result), rhs, rtol=1e-5, atol=1e-5):
                raise RuntimeError('Poisson problem could not be solved')
            logger.info('Solved Poisson problem with sparse.linalg.lsmr')
        
        # convert the result to the correct format
        if out is not None:
            out.flat[:] = result
        else:
            out = result.reshape(arr.shape)
        return out
    
    return solve_poisson



'''
Functions and classes for visualizing simulations.


.. autosummary::
   :nosignatures:

   movies
   plotting
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from .plotting import plot_magnitudes, plot_kymograph
from .movies import movie_scalar, movie_multiple



'''
Defines a collection of fields to represent multiple fields defined on a common
grid.

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import json
from typing import (Sequence, Optional, Union, Any, Dict,
                    List, Iterator)  # @UnusedImport

import numpy as np

from .base import FieldBase, DataFieldBase, OptionalArrayLike
from .scalar import ScalarField
from ..tools.plotting import plot_on_figure, PlotReference
from ..grids.base import GridBase



class FieldCollection(FieldBase):
    """ Collection of fields defined on the same grid """


    def __init__(self, fields: Sequence[DataFieldBase],
                 data: OptionalArrayLike = None,
                 copy_fields: bool = False,
                 label: Optional[str] = None):
        """ 
        Args:
            fields:
                Sequence of the individual fields
            data (:class:`numpy.ndarray`):
                Data of the fields. If `None`, the data is instead taken from
                the individual fields given by `fields`.
            copy_fields (bool):
                Flag determining whether the individual fields given in `fields`
                are copied. 
            label (str):
                Label of the field collection
            
        Warning:
            If `data` is given and `copy_fields == False`, the data in the
            individual fields is overwritten by the associated `data`.
        """
        if isinstance(fields, FieldCollection):
            # support assigning a field collection for convenience
            fields = fields.fields
            
        if len(fields) == 0:
            raise ValueError('At least one field must be defined')
        
        # check if grids are compatible
        grid = fields[0].grid
        if any(grid != f.grid for f in fields[1:]):
            grids = [f.grid for f in fields]
            raise RuntimeError(f'Grids are incompatible: {grids}')

        # create the list of underlying fields        
        if copy_fields:
            self._fields = [field.copy() for field in fields]
        else:
            self._fields = fields  # type: ignore
        
        # extract data from individual fields
        new_data: List[np.ndarray] = []
        self._slices: List[slice] = []
        dof = 0  # count local degrees of freedom
        for field in self.fields:
            if not isinstance(field, DataFieldBase):
                raise RuntimeError('Individual fields must be of type '
                                   'DataFieldBase. Field collections cannot be '
                                   'nested')
            start = len(new_data)
            this_data = field._data_flat
            new_data.extend(this_data)
            self._slices.append(slice(start, len(new_data)))
            dof += len(this_data)
                
        # combine into one data field
        data_shape = (dof,) + grid.shape
        if data is None:
            data_arr = np.array(new_data, dtype=np.double)
        else:
            data_arr = np.asarray(data, dtype=np.double)
            if data_arr.shape != data_shape:
                data_arr = np.array(np.broadcast_to(data_arr, data_shape))
        assert data_arr.shape == data_shape
        
        # initialize the class
        super().__init__(grid, data_arr, label=label)        
            
        # link the data of the original fields back to self._data if they were
        # not copied
        if not copy_fields:
            for i, field in enumerate(self.fields):
                field_shape = field.data.shape
                field._data_flat = self.data[self._slices[i]]
                
                # check whether the field data is based on our data field
                assert field.data.base is self.data
                assert field.data.shape == field_shape
         
         
    def __repr__(self):
        """ return instance as string """
        fields = []
        for f in self.fields:
            name = f.__class__.__name__
            if f.label:
                fields.append(f'{name}(..., label="{f.label}")')
            else:
                fields.append(f'{name}(...)')
        return f"{self.__class__.__name__}({', '.join(fields)})"

        
    def __len__(self):
        """ return the number of stored fields """
        return len(self.fields)
    
    
    def __iter__(self) -> Iterator[DataFieldBase]:
        """ return iterator over the actual fields """
        return iter(self.fields)
    
    
    def __getitem__(self, index: Union[int, str]) -> DataFieldBase:
        """ return a specific field """
        if isinstance(index, int):
            # simple numerical index
            return self.fields[index]
        
        elif isinstance(index, str):
            # index specifying the label of the field
            for field in self.fields:
                if field.label == index:
                    return field
            raise KeyError(f'No field with name {index}')
        
        else:
            raise TypeError(f'Unsupported index {index}')

        
    def __setitem__(self, index: int, value):
        """ set the value of a specific field """
        # We need to load the field and set data explicitly
        # WARNING: Do not use `self.fields[index] = value`, since this would
        # break the connection between the data fields 
        if isinstance(index, int):
            # simple numerical index
            self.fields[index].data = value
            
        elif isinstance(index, str):
            # index specifying the label of the field
            for field in self.fields:
                if field.label == index:
                    field.data = value
                    break
            else:
                raise KeyError(f'No field with name {index}')

        else:
            raise TypeError(f'Unsupported index {index}')


    @property
    def fields(self) -> List[DataFieldBase]:
        """ list: the fields of this collection """
        return self._fields

        
    def __eq__(self, other):
        """ test fields for equality, ignoring the label """
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.fields == other.fields
    
        
    @classmethod
    def from_state(cls, attributes: Dict[str, Any],
                   data: np.ndarray = None) -> "FieldCollection":
        """ create a field collection from given state.
        
        Args:
            attributes (dict):
                The attributes that describe the current instance
            data (:class:`numpy.ndarray`, optional):
                Data values at support points of the grid defining all fields
        """
        if 'class' in attributes:
            class_name = attributes.pop('class')
            assert class_name == cls.__name__
        
        # restore the individual fields (without data)
        fields = [FieldBase.from_state(field_state)
                  for field_state in attributes.pop('fields')]
        
        return cls(fields, data=data, **attributes)  # type: ignore


    @classmethod
    def _from_hdf_dataset(cls, dataset) -> "FieldCollection":
        """ construct the class by reading data from an hdf5 dataset """
        # copy attributes from hdf
        attributes = dict(dataset.attrs)
        
        # determine class
        class_name = json.loads(attributes.pop('class'))
        assert class_name == cls.__name__
        
        # determine the fields
        field_attrs = json.loads(attributes.pop('fields'))
        fields = [DataFieldBase._from_hdf_dataset(dataset[f"field_{i}"])
                  for i in range(len(field_attrs))]
        
        # unserialize remaining attributes
        attributes = cls.unserialize_attributes(attributes)
        return cls(fields, **attributes)  # type: ignore


    def _write_hdf_dataset(self, hdf_path):
        """ write data to a given hdf5 path `hdf_path` """
        # write attributes of the collection
        for key, value in self.attributes_serialized.items():
            hdf_path.attrs[key] = value
                  
        # write individual fields
        for i, field in enumerate(self.fields):
            field._write_hdf_dataset(hdf_path, f"field_{i}")


    def assert_field_compatible(self, other: FieldBase,
                                accept_scalar: bool = False):
        """ checks whether `other` is compatible with the current field
        
        Args:
            other (FieldBase): Other field this is compared to
            accept_scalar (bool, optional): Determines whether it is acceptable
                that `other` is an instance of
                :class:`~pde.fields.ScalarField`.
        """
        super().assert_field_compatible(other, accept_scalar=accept_scalar)

        # check whether all sub fields are compatible
        if isinstance(other, FieldCollection):
            for f1, f2 in zip(self, other):
                f1.assert_field_compatible(f2, accept_scalar=accept_scalar)
                

    @classmethod
    def scalar_random_uniform(cls, num_fields: int, grid: GridBase,
                              vmin: float = 0, vmax: float = 1,
                              label: Optional[str] = None):
        """ create scalar fields with random values between `vmin` and `vmax`
        
        Args:
            num_fields (int): The number of fields to create
            grid (:class:`~pde.grids.GridBase`):
                Grid defining the space on which the fields are defined
            vmin (float): Smallest random value
            vmax (float): Largest random value
            label (str, optional): Name of the field collection
        """
        return cls([ScalarField.random_uniform(grid, vmin, vmax)
                    for _ in range(num_fields)], label=label)
    
    
    @property
    def attributes(self) -> Dict[str, Any]:
        """ dict: describes the state of the instance (without the data) """
        results = super().attributes
        del results['grid']
        results['fields'] = [f.attributes for f in self.fields]
        return results
    

    @property
    def attributes_serialized(self) -> Dict[str, str]:
        """ dict: serialized version of the attributes """
        results = {}
        for key, value in self.attributes.items():
            if key == 'fields':
                fields = [f.attributes_serialized for f in self.fields]
                results[key] = json.dumps(fields)
            else:
                results[key] = json.dumps(value)
        return results
    
    
    @classmethod
    def unserialize_attributes(cls, attributes: Dict[str, str]) \
            -> Dict[str, Any]:
        """ unserializes the given attributes
        
        Args:
            attributes (dict):
                The serialized attributes
                
        Returns:
            dict: The unserialized attributes
        """
        results = {}
        for key, value in attributes.items():
            if key == 'fields':
                results[key] = [FieldBase.unserialize_attributes(attrs)
                                for attrs in json.loads(value)]
            else:
                results[key] = json.loads(value)
        return results
    
    
    def copy(self, data: OptionalArrayLike = None, label: str = None) \
            -> 'FieldCollection':
        """ return a copy of the data, but not of the grid
        
        Args:
            data (:class:`numpy.ndarray`, optional):
                Data values at the support points of the grid that define the
                field. Note that the data is not copied but used directly.
            label (str, optional):
                Name of the copied field
        """
        if label is None:
            label = self.label
        fields = [f.copy() for f in self.fields]
        # if data is None, the data of the individual fields is copied in their
        # copy() method above. The underlying data is therefore independent from
        # the current field 
        return self.__class__(fields, data=data, label=label)


    def interpolate_to_grid(self, grid: GridBase,
                            method: str = 'numba',
                            fill: float = None,
                            label: Optional[str] = None) -> 'FieldCollection':
        """ interpolate the data of this field collection to another grid.
        
        Args:
            grid (:class:`~pde.grids.GridBase`):
                The grid of the new field onto which the current field is
                interpolated.
            method (str):
                Specifies interpolation method, e.g., 'numba', 'scipy_linear',
                'scipy_nearest' .
            fill (float, optional):
                Determines how values out of bounds are handled. If `None`, a
                `ValueError` is raised when out-of-bounds points are requested.
                Otherwise, the given value is returned.
            label (str, optional):
                Name of the returned field collection
            
        Returns:
            FieldCollection: Interpolated data
        """        
        if label is None:
            label = self.label
        fields = [f.interpolate_to_grid(grid, method=method, fill=fill)
                  for f in self.fields]
        return self.__class__(fields, label=label)        
        

    def smooth(self, sigma: Optional[float] = 1,
               out: Optional['FieldCollection'] = None,
               label: str = None) -> 'FieldCollection':
        """ applies Gaussian smoothing with the given standard deviation

        This function respects periodic boundary conditions of the underlying
        grid, using reflection when no periodicity is specified.
        
        sigma (float, optional):
            Gives the standard deviation of the smoothing in real length units
            (default: 1)
        out (FieldCollection, optional):
            Optional field into which the smoothed data is stored
        label (str, optional):
            Name of the returned field

        Returns:
            Field collection with smoothed data, stored at `out` if given.             
        """
        # allocate memory for storing output
        if out is None:
            out = self.copy(label=label)
        else:
            self.assert_field_compatible(out)
            if label:
                out.label = label
         
        # apply Gaussian smoothing for each axis
        for f_in, f_out in zip(self, out):        
            f_in.smooth(sigma=sigma, out=f_out)
             
        return out   

        
    def get_line_data(self,  # type: ignore
                      index: int = 0,
                      scalar: str = 'auto',
                      extract: str = 'auto') -> Dict[str, Any]:
        r""" return data for a line plot of the field
        
        Args:
            index (int):
                Index of the field whose data is returned
            scalar (str or int):
                The method for extracting scalars as described in
                :meth:`DataFieldBase.to_scalar`.
            extract (str):
                The method used for extracting the line data. See the docstring
                of the grid method `get_line_data` to find supported values.
                
        Returns:
            dict: Information useful for performing a line plot of the field
        """
        return self[index].get_line_data(scalar=scalar, extract=extract)
    
    
    def get_image_data(self, index: int = 0, **kwargs) -> Dict[str, Any]:
        r""" return data for plotting an image of the field

        Args:
            index (int): Index of the field whose data is returned
            \**kwargs: Arguments forwarded to the `get_image_data` method
                
        Returns:
            dict: Information useful for plotting an image of the field
        """
        return self[index].get_image_data(**kwargs)


    def _update_plot(self, reference: List[PlotReference]) -> None:
        """ update a plot collection with the current field values
        
        Args:
            reference (list of :class:`PlotReference`):
                All references of the plot to update
        """
        for field, ref in zip(self.fields, reference):
            field._update_plot(ref)
        
        
    @plot_on_figure(update_method='_update_plot')
    def plot(self,
             fig,
             kind: str = 'auto',
             resize_fig: bool = True,
             **kwargs) -> List[PlotReference]:
        r""" visualize all the fields in the collection
        
        Args:
            fig (:class:`matplotlib.figures.Figure`):
                Figure that is used for plotting. If omitted, a new figure is
                created.
            kind (str):
                Determines the visualizations. Supported values are `image`, 
                `line`, `vector`, or `interactive`. Alternatively, `auto`
                determines the best visualization based on each field itself.
            resize_fig (bool):
                Whether to resize the figure
            \**kwargs:
                All additional keyword arguments are forwarded to the actual
                plotting function.
                
        Returns:
            list of :class:`PlotReference`: Instances that contain information
            to update all the plots with new data later.
        """
        # disable interactive plotting temporarily
        # create a plot with all the panels 
        if resize_fig:
            fig.set_size_inches((4 * len(self), 3), forward=True)
        axs = fig.subplots(1, len(self))
        
        # plot all the elements into the respective axes
        reference = [field.plot(kind=kind, ax=ax, show=False, **kwargs)
                     for field, ax in zip(self.fields, axs)]
                
        # return the references for all subplots
        return reference
    
            
    def plot_interactive(self, scalar: str = 'auto', **kwargs):
        """ create an interactive plot of the field using :mod:`napari`

        Args:
            scalar (str): The method for obtaining scalar values of fields        
            **kwargs: Extra arguments are passed to :class:`napari.Viewer`
        """
        from ..tools.plotting import napari_viewer
        with napari_viewer(self.grid, **kwargs) as viewer:
            for field in self:
                viewer.add_image(field.to_scalar(scalar).data,
                                 name=field.label,
                                 rgb=False,
                                 scale=self.grid.discretization)



r"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>

This module handles the boundaries of all axes of a grid. It only defines
:class:`Boundaries`, which acts as a list of
:class:`~pde.grids.boundaries.axis.BoundaryAxisBase`.
"""

from typing import Union, Sequence
    
import numpy as np

from .axis import BoundaryPair, BoundaryPairData, get_boundary_axis
from ..base import GridBase
from .. import PeriodicityError
 
BoundariesData = Union[BoundaryPairData,
                       Sequence[BoundaryPairData]]



class Boundaries(list):
    """ class that bundles all boundary conditions for all axes """
    
    
    grid: GridBase
    """ :class:`~pde.grids.base.GridBase`:
    The grid for which the boundaries are defined """
    
    
    def __init__(self, boundaries):
        """ initialize with a list of boundaries """
        if len(boundaries) == 0:
            raise ValueError("List of boundaries must not be empty")
        
        # extract grid 
        self.grid = boundaries[0].grid
        
        # check dimension
        if len(boundaries) != self.grid.num_axes:
            raise ValueError('Need boundary conditions for '
                             f'{self.grid.num_axes} axes.')

        # check consistency
        for axis, boundary in enumerate(boundaries):
            if boundary.grid != self.grid:
                raise ValueError('Boundaries are not defined on the same grid')
            if boundary.axis != axis:
                raise ValueError('The boundaries need to be ordered according '
                                 'to the axes they apply to')
            if boundary.periodic != self.grid.periodic[axis]:
                raise PeriodicityError('Periodicity specified in the '
                                       'boundaries conditions is not '
                                       'compatible with the grid')
        
        # create the list of boundaries
        super().__init__(boundaries)
        
    
    def __str__(self):
        items = ', '.join(str(item) for item in self)
        return f'[{items}]'

    
    @classmethod
    def from_data(cls, grid: GridBase, boundaries, rank: int = 0) \
            -> "Boundaries":
        """
        Creates all boundaries from given data
        
        Args:
            grid (:class:`~pde.grids.base.GridBase`):
                The grid with which the boundary condition is associated
            boundaries (str or list or tuple or dict):
                Data that describes the boundaries. This can either be a list of
                specifications for each dimension or a single one, which is then
                applied to all dimensions. The boundary for a dimensions can be
                specified by one of the following formats:
                
                * string specifying a single type for all boundaries
                * dictionary specifying the type and values for all boundaries
                * tuple pair specifying the low and high boundary individually
            rank (int):
                The tensorial rank of the value associated with the boundary
                conditions.
                
        """
        # check whether this is already the correct class
        if isinstance(boundaries, Boundaries):
            # boundaries are already in the correct format
            assert boundaries.grid == grid
            boundaries.check_value_rank(rank)
            return boundaries
        
        # convert natural boundary conditions if present
        if boundaries == 'natural':
            # set the respective natural conditions for all axes 
            boundaries = ['periodic' if periodic else 'no-flux'
                          for periodic in grid.periodic]
            
        elif hasattr(boundaries, '__iter__'):
            # convert natural boundary conditions on individual axes
            for i, boundary in enumerate(boundaries):
                if boundary == 'natural':
                    if grid.periodic[i]:
                        boundaries[i] = 'periodic'
                    else:
                        boundaries[i] = 'no-flux'
        
        # create the list of BoundaryAxis objects
        if isinstance(boundaries, (str, dict)):
            # one specification for all axes
            bcs = [get_boundary_axis(grid, i, boundaries, rank=rank)
                   for i in range(grid.num_axes)]
            
        elif len(boundaries) == grid.num_axes:
            # assume that data is given for each boundary
            bcs = [get_boundary_axis(grid, i, boundary, rank=rank)
                   for i, boundary in enumerate(boundaries)]
            
        elif grid.num_axes == 1 and len(boundaries) == 2:
            # special case where the two sides can be specified directly
            bcs = [get_boundary_axis(grid, 0, boundaries, rank=rank)]
            
        else:
            raise ValueError(f'Unsupported boundary format: `{boundaries}`. '
                             f'{cls.get_help()}')
            
        return cls(bcs)
    
    
    def __eq__(self, other):
        if not isinstance(other, Boundaries):
            return NotImplemented
        return super().__eq__(other) and self.grid == other.grid
    
    
    def _cache_hash(self) -> int:
        """ returns a value to determine when a cache needs to be updated """ 
        return hash(tuple(bc_ax._cache_hash() for bc_ax in self))
    
    
    def check_value_rank(self, rank: int):
        """ check whether the values at the boundaries have the correct rank
        
        Args:
            rank (tuple): The rank of the value that is stored with this
                boundary condition
            
        Throws:
            RuntimeError: if any value does not have rank `rank`
        """
        for b in self:
            b.check_value_rank(rank)


    @classmethod
    def get_help(cls) -> str:
        """ Return information on how boundary conditions can be set """
        return ("Boundary conditions for each axis are set using a list: "
                "[bc_x, bc_y, bc_z]. If the associated axis is periodic, the "
                "boundary condition needs to be set to 'periodic'. Otherwise, "
                f"{BoundaryPair.get_help()}")       


    def copy(self, value=None) -> "Boundaries":
        """ create a copy of the current boundaries
        
        Args:
            value (float or array, optional):
                If given, this changes the value stored with the boundary
                conditions. The interpretation of this value depends on the type
                of boundary condition.
            copy_grid (bool):
                Whether the grid should also be copied
        """
        result = self.__class__([bc.copy() for bc in self])
        if value is not None:
            result.set_value(value)
        return result
    
    
    @property
    def periodic(self) -> np.ndarray:
        """ :class:`numpy.ndarray`: a boolean array indicating which dimensions
        are periodic according to the boundary conditions """
        return self.grid.periodic


    def set_value(self, value=0):
        """ set the value of all non-periodic boundaries
        
        Args:
            value (float or array):
                Sets the value stored with the boundary conditions. The
                interpretation of this value depends on the type of boundary
                condition.
        """
        for b in self:
            if not b.periodic:
                b.set_value(value)

        
    @property
    def _scipy_border_mode(self) -> dict:
        """ dict: return a dictionary that can be used in the scipy ndimage
        functions to specify the border mode. If the current boundary cannot be
        represented by these modes, a RuntimeError is raised
        """
        mode: dict = self[0]._scipy_border_mode
        for b in self[1:]:
            if mode != b._scipy_border_mode:
                raise RuntimeError('Incompatible dimensions')
        return mode
    
    
    @property
    def _uniform_discretization(self) -> float:
        """ float: returns the uniform discretization or raises RuntimeError """
        dx_mean = np.mean(self.grid.discretization)
        if np.allclose(self.grid.discretization, dx_mean):
            return float(dx_mean)
        else:
            raise RuntimeError('Grid discretization is not uniform')
    
            
    def extract_component(self, *indices) -> "Boundaries":
        """ extracts the boundary conditions of the given extract_component.

        Args:
            *indices:
                One or two indices for vector or tensor fields, respectively
        """
        boundaries = [boundary.extract_component(*indices)
                      for boundary in self]
        return self.__class__(boundaries)
        
        
    @property
    def differentiated(self) -> "Boundaries":
        """ Domain: with differentiated versions of all boundary conditions """
        return self.__class__([b.differentiated for b in self])
        
    


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from .. import movies
from ...grids import UnitGrid
from ...fields import ScalarField
from ...storage import MemoryStorage
from ...pdes import DiffusionPDE
from ...tools.misc import skipUnlessModule



@skipUnlessModule("matplotlib")
def test_movie(tmp_path):
    """ test Movie class"""
    import matplotlib.pyplot as plt
    
    with movies.Movie(verbose=False) as movie:
        # iterate over all time steps
        plt.plot([0, 1], [0, 1])
        movie.add_figure()
        movie.add_figure()
        
        folder = tmp_path / 'test_movie' 
        folder.mkdir(exist_ok=True)
        movie.save_frames(folder / "frame_%09d.png")
        num_files = sum(1
                        for e in folder.iterdir()
                        if e.is_file() and not e.name.startswith('.'))
        assert num_files == 2
    
        # save movie
        path = tmp_path / "test_movie.mov"
        try:
            movie.save(path)
        except FileNotFoundError:
            pass  # can happen when ffmpeg is not installed
        else:
            assert path.stat().st_size > 0



@skipUnlessModule("matplotlib")
def test_movie_scalar():
    """ test Movie class"""
    
    # create some data
    state = ScalarField.random_uniform(UnitGrid([16, 16]))
    pde = DiffusionPDE()
    storage = MemoryStorage()
    tracker = storage.tracker(interval=1)
    pde.solve(state, t_range=2, dt=1e-2, backend='numpy', tracker=tracker)
    
    # check creating the movie
    movies.movie_scalar(storage, filename=None, progress=False)





"""
Setting boundary conditions
===========================

This example shows how different boundary conditions can be specified.
"""

from pde import UnitGrid, ScalarField, DiffusionPDE

grid = UnitGrid([16, 16], periodic=[False, True]) # generate grid
state = ScalarField.random_uniform(grid, 0.2, 0.3)  # generate initial condition

# set boundary conditions `bc` for all axes
bc_x_left = {'derivative': 0.1}  
bc_x_right = {'value': 'sin(y)'}
bc_x = [bc_x_left, bc_x_right]
bc_y = 'periodic' 
eq = DiffusionPDE(bc=[bc_x, bc_y])

result = eq.solve(state, t_range=10, dt=0.005)
result.plot(show=True)



"""
Defines a class storing data in memory. 
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import Optional, List, Sequence
from contextlib import contextmanager

import numpy as np

from .base import StorageBase, InfoDict
from ..fields import FieldCollection
from ..fields.base import FieldBase 

    
    
class MemoryStorage(StorageBase):
    """ simple storage in memory """
    
    
    def __init__(self, times: Optional[List[float]] = None,
                 fields: Optional[Sequence[FieldBase]] = None,
                 info: InfoDict = None,
                 write_mode: str = 'truncate_once'):
        """
        Args:
            times (:class:`numpy.ndarray`):
                Sequence of times for which data is known
            fields (list of :class:`~pde.fields.FieldBase`):
                The field data at the given times
            info (dict):
                Supplies extra information that is stored in the storage
            write_mode (str):
                Determines how new data is added to already existing data.
                Possible values are: 'append' (data is always appended),
                'truncate' (data is cleared every time this storage is used
                for writing), or 'truncate_once' (data is cleared for the first
                writing, but appended subsequently). Alternatively, specifying
                'readonly' will disable writing completely.
        """
        super().__init__(info=info, write_mode=write_mode)
        self.times: List[float] = [] if times is None else times
        
        if fields is None:
            self._grid = None
            self.data = []
        else:
            self._grid = fields[0].grid
            self.data = [fields[0].data]
            for field in fields[1:]:
                if self._grid != field.grid:
                    raise ValueError('Grids of the fields are incompatible')
                self.data.append(field.data)
            self._data_shape = self.data[0].shape
            
        # check consistency
        if len(self.times) != len(self.data):
            raise ValueError('Length of the supplied `times` and `fields` are '
                             f'inconsistent ({len(self.times)} != '
                             f'{len(self.data)}).')
            

    @classmethod            
    def from_collection(cls, storages: Sequence["StorageBase"],
                        label: str = None) -> "MemoryStorage":
        """ combine multiple memory storages into one
        
        Args:
            storages (list): 
                A collection of instances of
                :class:`~pde.storage.base.StorageBase` whose data
                will be concatenated into a single MemoryStorage
            label (str, optional):
                The label of the instances of
                :class:`~pde.fields.FieldCollection` that
                represent the concatenated data
                
        Returns:
            :class:`~pde.storage.memory.MemoryStorage`: Storage
            containing all the data.
        """
        if len(storages) == 0:
            return cls()
        
        # initialize the combined data
        times = storages[0].times
        data = [[field] for field in storages[0]]
        
        # append data from further storages
        for storage in storages[1:]:
            if not np.allclose(times, storage.times):
                raise ValueError('Storages have incompatible times')
            for i, field in enumerate(storage):
                data[i].append(field)
                
        # convert data format to FieldCollections
        fields = [FieldCollection(d, label=label)  # type: ignore
                  for d in data]
        
        return cls(times, fields=fields)  # type: ignore
            
        
    def clear(self, clear_data_shape: bool = False) -> None:
        """ truncate the storage by removing all stored data.
        
        Args:
            clear_data_shape (bool): Flag determining whether the data shape is
                also deleted.
        """
        self.times = []
        self.data = []
        super().clear(clear_data_shape=clear_data_shape)


    def start_writing(self, field: FieldBase, info: InfoDict = None) -> None:
        """ initialize the storage for writing data
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                An example of the data that will be written to extract the grid
                and the data_shape
            info (dict):
                Supplies extra information that is stored in the storage
        """
        super().start_writing(field, info=info)
        
        # update info after opening file because otherwise information can be
        # overwritten by data that is already present in the file
        if info is not None:
            self.info.update(info)
            
        # handle the different write modes
        if self.write_mode == 'truncate_once':
            self.clear()
            self.write_mode = 'append'  # do not truncate in subsequent calls
            
        elif self.write_mode == 'truncate':
            self.clear()

        elif self.write_mode == 'readonly':
            raise RuntimeError('Cannot write in read-only mode')
        
        elif self.write_mode != 'append':        
            raise ValueError(f'Unknown write mode `{self.write_mode}`. '
                             'Possible values are `truncate_once`, '
                             '`truncate`, and `append`')
                    
            
    def append(self, data: np.ndarray, time: Optional[float] = None) -> None:
        """ append a new data set
        
        Args:
            data (:class:`numpy.ndarray`): The actual data
            time (float, optional): The time point associated with the data
        """
        assert data.shape == self.data_shape
        self.data.append(np.array(data))  # store copy of the data
        if time is None:
            time = 0 if len(self.times) == 0 else self.times[-1] + 1
        self.times.append(time)



@contextmanager
def get_memory_storage(field: FieldBase, info: InfoDict = None):
    """ initialize the storage for writing data
    
    Args:
        field (:class:`~pde.fields.FieldBase`):
            An example of the data that will be written to extract the grid
            and the data_shape
        info (dict):
            Supplies extra information that is stored in the storage
    """
    storage = MemoryStorage()
    storage.start_writing(field, info)
    yield storage
    storage.end_writing()



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from .. import plotting
from ...grids import UnitGrid
from ...fields import ScalarField, FieldCollection
from ...storage import get_memory_storage
from ...tools.misc import skipUnlessModule



@skipUnlessModule("matplotlib")
def test_scalar_field_plot(tmp_path):
    """ test ScalarFieldPlot class"""
    path = tmp_path / "test_scalar_field_plot.png"
            
    # create some data
    state = ScalarField.random_uniform(UnitGrid([16, 16]))
    for scale in [(0, 1), 1, 'automatic', 'symmetric', 'unity']:
        sfp = plotting.ScalarFieldPlot(state, scale=scale)
        sfp.savefig(path)
        assert path.stat().st_size > 0

    sfp = plotting.ScalarFieldPlot(state, quantities={'source': None})
    sfp.savefig(path)
    assert path.stat().st_size > 0



@skipUnlessModule("matplotlib")
def test_scalar_plot(tmp_path):
    """ test Simple simulation """
    path = tmp_path / "test_scalar_plot.png"
    
    # create some data
    state = ScalarField.random_uniform(UnitGrid([16, 16]), label='test')
    with get_memory_storage(state) as storage:
        storage.append(state.data, 0)
        storage.append(state.data, 1)
    
    # check creating an overview image
    plotting.plot_magnitudes(storage, filename=path)
    assert path.stat().st_size > 0

    # check creating an kymograph
    plotting.plot_kymograph(storage, filename=path)
    assert path.stat().st_size > 0



@skipUnlessModule("matplotlib")
def test_collection_plot(tmp_path):
    """ test Simple simulation """
    # create some data
    field = FieldCollection([ScalarField(UnitGrid([8, 8]), label='first'),
                             ScalarField(UnitGrid([8, 8]))])
    with get_memory_storage(field) as storage:
        storage.append(field.data)
    
    path = tmp_path / 'test_collection_plot.png'
    plotting.plot_magnitudes(storage, filename=path)
    assert path.stat().st_size > 0




'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import pytest
import numpy as np

from ...fields import ScalarField
from ...grids import UnitGrid
from ...pdes import DiffusionPDE
from .. import Controller, ExplicitSolver



def test_compare_explicit():
    """ test explicit solvers """
    grid = UnitGrid([16, 16])
    field = ScalarField.random_uniform(grid, -1, 1)
    eq = DiffusionPDE()
    
    c1 = Controller(ExplicitSolver(eq), t_range=.1, tracker=None)
    s1 = c1.run(field, dt=2e-3)
    
    c2 = Controller(ExplicitSolver(eq, scheme='runge-kutta'), t_range=.1,
                    tracker=None)
    with np.errstate(under='ignore'):
        s2 = c2.run(field, dt=2e-3)
                               
    np.testing.assert_allclose(s1.data, s2.data, rtol=1e-2, atol=1e-2)



@pytest.mark.parametrize('backend', ['numba', 'numpy'])
def test_stochastic_solvers(backend):
    """ test simple version of the stochastic solver """
    field = ScalarField.random_uniform(UnitGrid([16]), -1, 1)
    eq = DiffusionPDE()
    seq = DiffusionPDE(noise=1e-6)
    
    solver1 = ExplicitSolver(eq, backend=backend)
    c1 = Controller(solver1, t_range=1, tracker=None)
    s1 = c1.run(field, dt=1e-3)
    
    solver2 = ExplicitSolver(seq, backend=backend)
    c2 = Controller(solver2, t_range=1, tracker=None)
    s2 = c2.run(field, dt=1e-3)
                               
    np.testing.assert_allclose(s1.data, s2.data, rtol=1e-4, atol=1e-4)    
    assert not solver1.info['stochastic']
    assert solver2.info['stochastic']
    


def test_unsupported_stochastic_solvers():
    """ test some solvers that do not support stochasticity """
    field = ScalarField.random_uniform(UnitGrid([16]), -1, 1)
    eq = DiffusionPDE(noise=1)
    
    with pytest.raises(RuntimeError):
        eq.solve(field, 1, method='explicit', scheme='runge-kutta')
    with pytest.raises(RuntimeError):
        eq.solve(field, 1, method='scipy', scheme='runge-kutta')



'''
Helper functions for just-in-time compilation with numba

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import logging
import os
import warnings
from functools import wraps
from typing import Callable, Dict, Any, Optional, Tuple, TypeVar

import numpy as np
import numba as nb  # lgtm [py/import-and-import-from]

from ..tools.misc import decorator_arguments


try:
    from numba.core.dispatcher import Dispatcher
except ImportError:
    # assume older numba module structure
    from numba.dispatcher import Dispatcher



# global settings for numba
NUMBA_PARALLEL = True   # enable parallel numba
NUMBA_FASTMATH = True   # enable fastmath switch (ignores NaNs!)
NUMBA_DEBUG = False     # enable additional debug information
    
    
# numba version as a list of integers
NUMBA_VERSION = [int(v) for v in nb.__version__.split('.')[:2]]
 
 
 
TFunc = TypeVar('TFunc')

    

def numba_environment() -> Dict[str, Any]:
    """ return information about the numba setup used
    
    Returns:
        (dict) information about the numba setup
    """
    # determine whether Nvidia Cuda is available
    try:
        from numba import cuda
        cuda_available = cuda.is_available()
    except ImportError:
        cuda_available = False

    # determine whether AMD ROC is available
    try:
        from numba import roc
        roc_available = roc.is_available()
    except ImportError:
        roc_available = False
    
    # determine threading layer
    try:
        threading_layer = nb.threading_layer()
    except ValueError:
        # threading layer was not initialized, so compile a mock function 
        @nb.jit('i8()', parallel=True)
        def f():
            s = 0
            for i in nb.prange(4):
                s += i
            return s
        f()
        try:
            threading_layer = nb.threading_layer()
        except ValueError:  # cannot initialize threading
            threading_layer = None
    except AttributeError:   # old numba version
        threading_layer = None
        
    return {'version': nb.__version__,
            'parallel': NUMBA_PARALLEL,
            'fastmath': NUMBA_FASTMATH,
            'debug': NUMBA_DEBUG,
            'using_svml': nb.config.USING_SVML,
            'threading_layer': threading_layer,
            'omp_num_threads': os.environ.get('OMP_NUM_THREADS'),
            'mkl_num_threads': os.environ.get('MKL_NUM_THREADS'),
            'num_threads': nb.config.NUMBA_NUM_THREADS,
            'num_threads_default': nb.config.NUMBA_DEFAULT_NUM_THREADS,
            'cuda_available': cuda_available,
            'roc_available': roc_available}
    
    
    
def _numba_get_signature(parallel: bool = False, **kwargs) -> Dict[str, Any]:
    """ return arguments for the :func:`nb.jit` with default values
    
    Args:
        parallel (bool): Allow parallel compilation of the function
        **kwargs: Additional arguments to `nb.jit`
        
    Returns:
        dict: Keyword arguments that can directly be used in :func:`nb.jit`
    """
    kwargs.setdefault('nopython', True)
    kwargs.setdefault('fastmath', NUMBA_FASTMATH)
    kwargs.setdefault('debug', NUMBA_DEBUG)
    
    # make sure parallel numba is only enabled in restricted cases 
    kwargs['parallel'] = parallel and NUMBA_PARALLEL
    return kwargs
    


if nb.config.DISABLE_JIT:    
    # use work-around for https://github.com/numba/numba/issues/4759
    def flat_idx(arr, i):
        """ helper function allowing indexing of scalars as if they arrays """
        if np.isscalar(arr):
            return arr
        else:
            return arr.flat[i]
else:
    # compiled version that specializes correctly
    @nb.generated_jit(nopython=True)
    def flat_idx(arr, i):
        """ helper function allowing indexing of scalars as if they arrays """
        if isinstance(arr, (nb.types.Integer, nb.types.Float)):
            return lambda arr, i: arr
        else:
            return lambda arr, i: arr.flat[i]

    
    
@decorator_arguments
def jit(function: TFunc,
        signature=None,
        parallel: bool = False, **kwargs) -> TFunc:
    """ apply nb.jit with predefined arguments
    
    Args:
        signature: Signature of the function to compile
        parallel (bool): Allow parallel compilation of the function
        **kwargs: Additional arguments to `nb.jit`
        
    Returns:
        Function that will be compiled using numba
    """
    if isinstance(function, Dispatcher):
        # function is already jited
        return function  # type: ignore
    
    jit_kwargs = _numba_get_signature(parallel, **kwargs)
    name = function.__name__  # type: ignore
    logging.getLogger(__name__).info('Compile `%s` with parallel=%s',
                                     name, jit_kwargs['parallel'])
    return nb.jit(signature, **jit_kwargs)(function)  # type: ignore


        
@decorator_arguments
def jit_allocate_out(func: Callable,
                     parallel: bool = False,
                     out_shape: Optional[Tuple[int, ...]] = None,
                     num_args: int = 1,
                     **kwargs) -> Callable:
    """ Decorator that compiles a function with allocating an output array. 
    
    This decorator compiles a function that takes the arguments `arr` and
    `out`. The point of this function is to make the `out` array optional by 
    supplying an empty array of the same shape as `arr` if necessary. This is
    implemented efficiently by using :func:`nb.generated_jit`. 
    
    Args:
        func: The function to be compiled
        parallel (bool): Determines whether the function is jitted with
            parallel=True.
        out_shape (tuple): Determines the shape of the `out` array. If omitted,
            the same shape as the input array is used.
        num_args (int, optional): Determines the number of input arguments of
            the function.
        **kwargs: Additional arguments to `nb.jit`
            
    Returns:
        The decorated function
    """
    # TODO: Remove `num_args` and use inspection on `func` instead
    
    # we need to cast `parallel` to bool since np.bool is not supported by jit
    parallel = bool(parallel)
    
    if nb.config.DISABLE_JIT:
        # jitting is disabled => return generic python function
        
        if num_args == 1:
            def f_arg1_with_allocated_out(arr, out=None):
                """ helper function allocating output array """
                if out is None:
                    if out_shape is None:
                        out = np.empty_like(arr)
                    else:
                        out = np.empty(out_shape)
                return func(arr, out)
            
            return f_arg1_with_allocated_out
            
        elif num_args == 2:
            def f_arg2_with_allocated_out(a, b, out=None):
                """ helper function allocating output array """
                if out is None:
                    assert a.shape == b.shape
                    if out_shape is None:
                        out = np.empty_like(a)
                    else:
                        out = np.empty(out_shape)
                return func(a, b, out)
            
            return f_arg2_with_allocated_out
            
        else:
            raise NotImplementedError('Only 1 or 2 arguments are supported')
            
        
    else:
        # jitting is enabled => return specific compiled functions
        jit_kwargs = _numba_get_signature(parallel=False, **kwargs)
        register_jitable = nb.extending.register_jitable
        logging.getLogger(__name__).info('Compile `%s` with parallel=%s',
                                         func.__name__, jit_kwargs['parallel'])
    
        if num_args == 1:
            @nb.generated_jit(**jit_kwargs)
            @wraps(func)
            def wrapper(arr, out=None):
                """ wrapper deciding whether the underlying function is called
                with or without `out`. This uses :func:`nb.generated_jit` to
                compile different versions of the same function
                """
                if isinstance(arr, nb.types.Number):
                    # simple scalar call -> do not need to allocate anything
                    raise RuntimeError('Functions defined with an `out` '
                                       'keyword should not be called with '
                                       'scalar quantities.')
                    
                if not isinstance(arr, nb.types.Array):
                    raise RuntimeError('Compiled functions need to be called '
                                       'with numpy arrays, not type '
                                       f'{arr.__class__.__name__}')
                
                if isinstance(out, (nb.types.NoneType, nb.types.Omitted)):
                    # function is called without `out`
                    f_jit = register_jitable(parallel=parallel)(func)
                    
                    if out_shape is None:
                        # we have to obtain the shape of `out` from `arr` 
                        def f_with_allocated_out(arr, out):
                            """ helper function allocating output array """
                            return f_jit(arr, out=np.empty_like(arr))
        
                    else:
                        # the shape of `out` is given by `out_shape` 
                        def f_with_allocated_out(arr, out):
                            """ helper function allocating output array """
                            return f_jit(arr, out=np.empty(out_shape))
        
                    return f_with_allocated_out
                
                else:
                    # function is called with `out` argument
                    return func    
            
        elif num_args == 2:
            @nb.generated_jit(**jit_kwargs)
            @wraps(func)
            def wrapper(a, b, out=None):
                """ wrapper deciding whether the underlying function is called
                with or without `out`. This uses nb.generated_jit to compile
                different versions of the same function. """
                if isinstance(a, nb.types.Number):
                    # simple scalar call -> do not need to allocate anything
                    raise RuntimeError('Functions defined with an `out` '
                                       'keyword should not be called with '
                                       'scalar quantities')
                
                elif isinstance(out, (nb.types.NoneType, nb.types.Omitted)):
                    # function is called without `out`
                    f_jit = register_jitable(parallel=parallel)(func)
                    
                    if out_shape is None:  
                        # we have to obtain the shape of `out` from `a` 
                        def f_with_allocated_out(a, b, out):
                            """ helper function allocating output array """
                            return f_jit(a, b, out=np.empty_like(a))
        
                    else:
                        # the shape of `out` is given by `out_shape` 
                        def f_with_allocated_out(a, b, out):
                            """ helper function allocating output array """
                            return f_jit(a, b, out=np.empty(out_shape))
        
                    return f_with_allocated_out
                
                else:
                    # function is called with `out` argument
                    return func   
                
        else:
            raise NotImplementedError('Only 1 or 2 arguments are supported')
                 
        return wrapper  # type: ignore
    


if nb.config.DISABLE_JIT:    
    # dummy function that creates a ctypes pointer
    def address_as_void_pointer(addr):
        """ returns a void pointer from a given memory address
        
        Example:
            This can for instance be used together with `numba.carray`:
            
            >>> addr = arr.ctypes.data
            >>> numba.carray(address_as_void_pointer(addr), arr.shape, arr.dtype
            
        Args:
            addr (int): The memory address
            
        Returns:
            :class:`ctypes.c_void_p`: Pointer to the memory address
        """
        import ctypes
        return ctypes.cast(addr, ctypes.c_void_p)
    
else:
    # actually useful function that creates a numba pointer
    @nb.extending.intrinsic
    def address_as_void_pointer(typingctx, src):
        """ returns a void pointer from a given memory address
        
        Example:
            This can for instance be used together with `numba.carray`:
            
            >>> addr = arr.ctypes.data
            >>> numba.carray(address_as_void_pointer(addr), arr.shape, arr.dtype
            
        Args:
            addr (int): The memory address
            
        Returns:
            :class:`numba.core.types.voidptr`: Pointer to the memory address
        """
        from numba.core import types, cgutils
        sig = types.voidptr(src)
        
        def codegen(cgctx, builder, sig, args):
            return builder.inttoptr(args[0], cgutils.voidptr_t)
        return sig, codegen



if NUMBA_VERSION < [0, 45]:
    warnings.warn('Your numba version is outdated. Please install at least '
                  'version 0.45.')



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import math
import random

import pytest
import numpy as np
from scipy import spatial, integrate

from .. import spherical



def test_volume_conversion():
    """ tests conversion of volume and radius of droplet """
    for dim in [1, 2, 3]:
        radius = 1 + random.random()
        volume = spherical.volume_from_radius(radius, dim=dim)
        radius2 = spherical.radius_from_volume(volume, dim=dim)
        assert radius2 == pytest.approx(radius)
        
        
        
def test_volume_conversion_numba():
    """ tests conversion of volume and radius of droplet using numba """
    for dim in [1, 2, 3]:
        r2v = spherical.make_volume_from_radius_compiled(dim)
        v2r = spherical.make_radius_from_volume_compiled(dim)
        radius = 1 + random.random()
        assert v2r(r2v(radius)) == pytest.approx(radius)
        
        
        
def test_surface():
    """ test whether the surface is calculated correctly """
    for dim in [1, 2, 3]:
        radius = 1 + random.random()
        eps = 1e-10
        vol1 = spherical.volume_from_radius(radius + eps, dim=dim)
        vol0 = spherical.volume_from_radius(radius, dim=dim)
        surface_approx = (vol1 - vol0) / eps
        surface = spherical.surface_from_radius(radius, dim=dim)
        assert surface == pytest.approx(surface_approx, rel=1e-3)
        
        r2s = spherical.make_surface_from_radius_compiled(dim)
        assert surface == pytest.approx(r2s(radius))
        
        if dim == 1:
            with pytest.raises(RuntimeError):
                spherical.radius_from_surface(surface, dim=dim)
        else:
            assert (spherical.radius_from_surface(surface, dim=dim) == 
                    pytest.approx(radius))


        
def test_spherical_conversion():
    """ test the conversion between spherical and Cartesian coordinates """
    s2c = spherical.points_spherical_to_cartesian
    c2s = spherical.points_cartesian_to_spherical
    
    ps = np.random.randn(64, 3)
    np.testing.assert_allclose(s2c(c2s(ps)), ps)
    
    # enforce angles
    ps[:, 0] = np.abs(ps[:, 0])  # radius is positive
    ps[:, 1] %= np.pi            # θ is between 0 and pi
    ps[:, 2] %= 2 * np.pi        # φ is between 0 and 2 pi
    np.testing.assert_allclose(c2s(s2c(ps)), ps)



def test_spherical_polygon_area():
    """ test the function get_spherical_polygon_area """
    area = spherical.get_spherical_polygon_area
    sector_area = area([[0, 1., 0], [0, 0, 1.], [-1., 0, 0]])
    assert sector_area == pytest.approx(np.pi / 2)

                   
                                
def test_spherical_voronoi():
    """ test spatial.SphericalVoronoi """
    # random points on the sphere
    ps = np.random.random((32, 3)) - 0.5
    ps /= np.linalg.norm(ps, axis=1)[:, None]

    voronoi = spatial.SphericalVoronoi(ps)
    voronoi.sort_vertices_of_regions()                

    total = sum(spherical.get_spherical_polygon_area(voronoi.vertices[reg])
                for reg in voronoi.regions)
    assert total == pytest.approx(4 * np.pi)
            
            
            
@pytest.mark.parametrize('dim', range(1, 4))
def test_points_on_sphere(dim, tmp_path):
    """ test spatial.SphericalVoronoi """
    shell = spherical.PointsOnSphere.make_uniform(dim=dim)
    assert shell.dim == dim
    
    for balance_axes in [True, False]:
        ws = shell.get_area_weights(balance_axes=balance_axes)
        assert ws.sum() == pytest.approx(1)
        np.testing.assert_allclose(ws, 1 / len(shell.points), rtol=0.1)
        
    ws = shell.get_area_weights(balance_axes=True)
    np.testing.assert_allclose(ws @ shell.points, 0, atol=1e-15)
    
    path = tmp_path / f'test_points_on_sphere_{dim}.xyz'
    shell.write_to_xyz(path=path)
    assert path.stat().st_size > 0
        
        
        
def test_points_on_sphere_2():
    """ special tests for 2 dimensions """
    num = np.random.randint(3, 9)
    shell = spherical.PointsOnSphere.make_uniform(dim=2, num_points=num)
    assert num * shell.get_mean_separation() == pytest.approx(2 * np.pi)
            
            
            
def test_spherical_index():
    """ test the conversion of the spherical index """
    # check conversion
    for k in range(20):
        l, m = spherical.spherical_index_lm(k)
        assert spherical.spherical_index_k(l, m) == k
        
    # check order
    k = 0
    for l in range(4):
        for m in range(-l, l+1):
            assert spherical.spherical_index_k(l, m) == k
            k += 1
            
    for l in range(4):
        assert (spherical.spherical_index_count(l) ==
                spherical.spherical_index_k(l, l) + 1)
        
    for l in range(4):
        for m in range(-l, l+1):
            k = spherical.spherical_index_k(l, m)
            assert spherical.spherical_index_count_optimal(k + 1) == (m == l)
            
          
            
def test_spherical_harmonics_real():
    """ test spherical harmonics """
    # test real spherical harmonics for symmetric case
    for deg in range(4):
        for _ in range(5):
            θ = math.pi * random.random()
            φ = 2 * math.pi * random.random()
            y1 = spherical.spherical_harmonic_symmetric(deg, θ)
            y2 = spherical.spherical_harmonic_real(deg, 0, θ, φ)
            assert y1 == y2
            
    # test orthogonality of real spherical harmonics
    deg = 1
    Ylm = spherical.spherical_harmonic_real
    for m1 in range(-deg, deg + 1):
        for m2 in range(-deg, m1 + 1):
            def integrand(t, p):
                return Ylm(deg, m1, t, p) * Ylm(deg, m2, t, p) * np.sin(t)
            overlap = integrate.dblquad(integrand, 0, 2 * np.pi,
                                        lambda _: 0, lambda _: np.pi)[0]
            if m1 == m2:
                assert overlap == pytest.approx(1)
            else:
                assert overlap == pytest.approx(0)



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np
import pytest

from .test_generic import iter_grids
from ..tensorial import Tensor2Field
from ..base import FieldBase
from ...grids import UnitGrid, CartesianGrid



def test_tensors():
    """ test some tensor calculations """
    grid = CartesianGrid([[0.1, 0.3], [-2, 3]], [3, 4])
    
    t1 = Tensor2Field(grid, np.full((2, 2) + grid.shape, 1))
    t2 = Tensor2Field(grid, np.full((2, 2) + grid.shape, 2))
    np.testing.assert_allclose(t2.average, [[2, 2], [2, 2]])
    assert t1.magnitude == pytest.approx(2)
    
    t3 = t1 + t2
    assert t3.grid == grid
    np.testing.assert_allclose(t3.data, 3)
    t1 += t2
    np.testing.assert_allclose(t1.data, 3)
    
    field = Tensor2Field.random_uniform(grid)
    trace = field.trace()
    from ..scalar import ScalarField
    assert isinstance(trace, ScalarField)
    np.testing.assert_allclose(trace.data, field.data.trace())
    
    t1 = Tensor2Field(grid)
    t1.data[0, 0, :] = 1
    t1.data[0, 1, :] = 2
    t1.data[1, 0, :] = 3
    t1.data[1, 1, :] = 4
    for method, value in [('min', 1), ('max', 4),
                          ('norm', np.linalg.norm([[1, 2], [3, 4]])),
                          ('squared_sum', 30), ('trace', 5),
                          ('invariant1', 5), ('invariant2', -1)]:
        p1 = t1.to_scalar(method)
        assert p1.data.shape == grid.shape
        np.testing.assert_allclose(p1.data, value)

    t2 = FieldBase.from_state(t1.attributes, data=t1.data)
    assert t1 == t2
    assert t1.grid is t2.grid
    
    attrs = Tensor2Field.unserialize_attributes(t1.attributes_serialized)
    t2 = FieldBase.from_state(attrs, data=t1.data)
    assert t1 == t2
    assert t1.grid is not t2.grid



def test_tensor_symmetrize():
    """ test advanced tensor calculations """
    grid = CartesianGrid([[0.1, 0.3], [-2, 3]], [2, 2])
    t1 = Tensor2Field(grid)
    t1.data[0, 0, :] = 1
    t1.data[0, 1, :] = 2
    t1.data[1, 0, :] = 3
    t1.data[1, 1, :] = 4
    
    # traceless = False
    t2 = t1.copy()
    t1.symmetrize(make_traceless=False, inplace=True)
    tr = t1.trace()
    assert np.all(tr.data == 5)
    t1_trans = np.swapaxes(t1.data, 0, 1)
    np.testing.assert_allclose(t1.data, t1_trans.data)
    
    ts = t1.copy()
    ts.symmetrize(make_traceless=False, inplace=True)
    np.testing.assert_allclose(t1.data, ts.data)
    
    # traceless = True
    t2.symmetrize(make_traceless=True, inplace=True)
    tr = t2.trace()
    assert np.all(tr.data == 0)
    t2_trans = np.swapaxes(t2.data, 0, 1)
    np.testing.assert_allclose(t2.data, t2_trans.data)

    ts = t2.copy()
    ts.symmetrize(make_traceless=True, inplace=True)
    np.testing.assert_allclose(t2.data, ts.data)
    
    

@pytest.mark.parametrize("grid", iter_grids())
def test_add_interpolated_tensor(grid):
    """ test the `add_interpolated` method """
    f = Tensor2Field(grid)
    a = np.random.random(f.data_shape)
    
    c = tuple(grid.point_to_cell(grid.get_random_point()))
    c_data = (Ellipsis,) + c
    p = grid.cell_to_point(c, cartesian=False)
    f.add_interpolated(p, a)
    np.testing.assert_almost_equal(f.data[c_data],
                                   a / grid.cell_volumes[c])
    
    f.add_interpolated(grid.get_random_point(cartesian=False), a)
    np.testing.assert_almost_equal(f.integral, 2 * a)
    
    f.data = 0  # reset
    add_interpolated = grid.make_add_interpolated_compiled()
    c = tuple(grid.point_to_cell(grid.get_random_point()))
    c_data = (Ellipsis,) + c
    p = grid.cell_to_point(c, cartesian=False)
    add_interpolated(f.data, p, a)
    np.testing.assert_almost_equal(f.data[c_data],
                                   a / grid.cell_volumes[c])

    add_interpolated(f.data, grid.get_random_point(cartesian=False), a)
    np.testing.assert_almost_equal(f.integral, 2 * a)



def test_tensor_invariants():
    """ test the invariants """
    # dim == 1
    f = Tensor2Field.random_uniform(UnitGrid([3]))
    np.testing.assert_allclose(f.to_scalar('invariant1').data,
                               f.to_scalar('invariant3').data)
    np.testing.assert_allclose(f.to_scalar('invariant2').data, 0)

    # dim == 2
    f = Tensor2Field.random_uniform(UnitGrid([3, 3]))
    invs = [f.to_scalar(f'invariant{i}').data for i in range(1, 4)]
    np.testing.assert_allclose(2 * invs[1], invs[2])

    a = np.random.uniform(0, 2 * np.pi)  # pick random rotation angle
    rot = Tensor2Field(f.grid)
    rot.data[0, 0, ...] = np.cos(a)
    rot.data[0, 1, ...] = np.sin(a)
    rot.data[1, 0, ...] = -np.sin(a)
    rot.data[1, 1, ...] = np.cos(a)
    f_rot = rot @ f @ rot.transpose()  # apply the transpose
    
    for i, inv in enumerate(invs, 1):
        np.testing.assert_allclose(inv, f_rot.to_scalar(f'invariant{i}').data,
                                   err_msg=f'Mismatch in invariant {i}')
        
    # dim == 3
    from scipy.spatial.transform import Rotation
    f = Tensor2Field.random_uniform(UnitGrid([1, 1, 1]))
    rot = Tensor2Field(f.grid)
    rot_mat = Rotation.from_rotvec(np.random.randn(3)).as_matrix()
    rot.data = rot_mat.reshape(3, 3, 1, 1, 1)
    f_rot = rot @ f @ rot.transpose()  # apply the transpose
    for i in range(1, 4):
        np.testing.assert_allclose(f.to_scalar(f'invariant{i}').data,
                                   f_rot.to_scalar(f'invariant{i}').data,
                                   err_msg=f'Mismatch in invariant {i}')
    


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np

from ...fields import ScalarField
from ...grids import UnitGrid
from ...pdes import DiffusionPDE
from .. import Controller, ScipySolver



def test_no_dt():
    """ test scipy solver without timestep """
    grid = UnitGrid([16])
    field = ScalarField.random_uniform(grid, -1, 1)
    eq = DiffusionPDE()
    
    c1 = Controller(ScipySolver(eq), t_range=1, tracker=None)
    s1 = c1.run(field, dt=1e-3)
    
    c2 = Controller(ScipySolver(eq), t_range=1, tracker=None)
    s2 = c2.run(field)
                               
    np.testing.assert_allclose(s1.data, s2.data, rtol=1e-3, atol=1e-3)




"""
Defines a solver using :mod:`scipy.integrate`
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import Callable

from scipy import integrate

from .base import SolverBase
from ..pdes.base import PDEBase
from ..fields.base import FieldBase



class ScipySolver(SolverBase):
    """ class for solving partial differential equations using scipy
    
    This class is a thin wrapper around :func:`scipy.integrate.solve_ivp`. In
    particular, it supports all the methods implemented by this function.
    """

    name = 'scipy'


    def __init__(self, pde: PDEBase, backend: str = 'auto', **kwargs):
        r""" initialize the explicit solver
        
        Args:
            pde (:class:`~pde.pdes.base.PDEBase`):
                The instance describing the pde that needs to be solved
            backend (str):
                Determines how the function is created. Accepted  values are
                'numpy` and 'numba'. Alternatively, 'auto' lets the code decide
                for the most optimal backend.
            **kwargs:
                All extra arguments are forwarded to 
                :func:`scipy.integrate.solve_ivp`.
        """        
        super().__init__(pde)
        self.backend = backend
        self.solver_params = kwargs
    

    def make_stepper(self, state: FieldBase, dt: float = None) -> Callable:
        """ return a stepper function
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other
                information can be extracted
            dt (float):
                Initial time step for the simulation. If `None`, the solver will
                choose a suitable initial value
                
        Returns:
            Function that can be called to advance the `state` from time
            `t_start` to time `t_end`. The function call signature is
            `(state: numpy.ndarray, t_start: float, t_end: float)`        
        """
        shape = state.data.shape
        self.info['dt'] = dt
        self.info['steps'] = 0    
        self.info['stochastic'] = False
        
        # obtain function for evaluating the right hand side
        rhs = self._make_pde_rhs(state, backend=self.backend,
                                 allow_stochastic=False)
        
        def rhs_helper(t, state_flat):
            """ helper function to provide the correct call convention """
            return rhs(state_flat.reshape(shape), t).flat
        
        def stepper(state, t_start, t_end):
            """ use scipy.integrate.odeint to advance `state` from `t_start` to
            `t_end` """
            if dt is not None:
                self.solver_params['first_step'] = min(t_end - t_start, dt)
                
            sol = integrate.solve_ivp(rhs_helper, 
                                      t_span=(t_start, t_end),
                                      y0=state.data.flat,
                                      t_eval=[t_end],  # only store necessary
                                      **self.solver_params)
            self.info['steps'] += sol.nfev
            state.data.flat = sol.y
            return sol.t[0]
            
        if dt:
            self._logger.info(f'Initialized {self.__class__.__name__} stepper '
                              'with dt=%g', dt)
        else:
            self._logger.info(f'Initialized {self.__class__.__name__} stepper')
        return stepper

        


r"""
Custom Class for coupled PDEs
=============================

This example shows how to solve a set of coupled PDEs, the
spatially coupled `FitzHugh–Nagumo model 
<https://en.wikipedia.org/wiki/FitzHugh–Nagumo_model>`_, which is a simple model
for the excitable dynamics of coupled Neurons:

.. math::

    \partial_t u &= \nabla^2 u + u (u - \alpha) (1 - u) + w \\
    \partial_t w &= \epsilon u
    
Here, :math:`\alpha` denotes the external stimulus and :math:`\epsilon` defines
the recovery time scale. We implement this as a custom PDE class below.
"""

from pde import UnitGrid, FieldCollection, PDEBase


class FitzhughNagumoPDE(PDEBase):
    """ FitzHugh–Nagumo model with diffusive coupling """
    
    def __init__(self, stimulus=.1, ε=1, bc='natural'):
        self.bc = bc
        self.stimulus = stimulus
        self.ε = ε
                
    def evolution_rate(self, state, t=0):
        u, w = state  # membrane potential and recovery variable
        
        u_t = u.laplace(bc=self.bc) + u * (u - self.stimulus) * (1 - u) + w
        w_t = self.ε * u
        
        return FieldCollection([u_t, w_t])


grid = UnitGrid([16, 16])
state = FieldCollection.scalar_random_uniform(2, grid)

eq = FitzhughNagumoPDE()
result = eq.solve(state, t_range=10, dt=0.01)
result.plot(show=True)



"""
The Kardar–Parisi–Zhang (KPZ) equation describing the evolution of an interface

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import Callable

import numpy as np


from .base import PDEBase
from ..fields import ScalarField
from ..grids.boundaries.axes import BoundariesData
from ..tools.numba import nb, jit
from ..tools.docstrings import fill_in_docstring


        
class KuramotoSivashinskyPDE(PDEBase):
    r""" The Kuramoto-Sivashinsky equation
    
    The mathematical definition is
    
    .. math::
        \partial_t u = -\nu \nabla^4 u  - \nabla^2 u -
            \frac{1}{2} \left(\nabla h\right)^2  + \eta(\boldsymbol r, t)
        
    where :math:`u` is the height of the interface in Monge parameterization.
    The dynamics are governed by the parameters :math:`\nu` , while
    :math:`\eta` is Gaussian white noise, whose strength is controlled by the
    `noise` argument.
    """

    explicit_time_dependence = False
    

    @fill_in_docstring
    def __init__(self, nu: float = 1,
                 noise: float = 0,
                 bc: BoundariesData = 'natural',
                 bc_lap: BoundariesData = None):
        r""" 
        Args:
            nu (float):
                Parameter :math:`\nu` for the strength of the fourth-order term
            noise (float):
                Strength of the (additive) noise term
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            bc_lap:
                The boundary conditions applied to the second derivative of the
                scalar field :math:`c`. If `None`, the same boundary condition
                as `bc` is chosen. Otherwise, this supports the same options as
                `bc`.
        """
        super().__init__(noise=noise)
        
        self.nu = nu
        self.bc = bc
        self.bc_lap = bc if bc_lap is None else bc_lap
            
            
    def evolution_rate(self, state: ScalarField,  # type: ignore
                       t: float = 0) -> ScalarField:
        """ evaluate the right hand side of the PDE
        
        Args:
            state (:class:`~pde.fields.ScalarField`):
                The scalar field describing the concentration distribution
            t (float): The current time point
            
        Returns:
            :class:`~pde.fields.ScalarField`:
            Scalar field describing the evolution rate of the PDE 
        """
        assert isinstance(state, ScalarField)
        state_lap = state.laplace(bc=self.bc)
        result = -self.nu * state_lap.laplace(bc=self.bc_lap) - state_lap \
                 - 0.5 * state.gradient(bc=self.bc).to_scalar('squared_sum')
        result.label = 'evolution rate'
        return result  # type: ignore
    
    
    def _make_pde_rhs_numba(self, state: ScalarField  # type: ignore
                            ) -> Callable:
        """ create a compiled function evaluating the right hand side of the PDE
        
        Args:
            state (:class:`~pde.fields.ScalarField`):
                An example for the state defining the grid and data types
                
        Returns:
            A function with signature `(state_data, t)`, which can be called
            with an instance of :class:`numpy.ndarray` of the state data and
            the time to obtained an instance of :class:`numpy.ndarray` giving
            the evolution rate.  
        """
        shape = state.grid.shape
        dim = state.grid.dim
        arr_type = nb.typeof(np.empty(shape, dtype=np.double))
        signature = arr_type(arr_type, nb.double)
        
        nu_value = self.nu
        laplace = state.grid.get_operator('laplace', bc=self.bc)
        laplace2 = state.grid.get_operator('laplace', bc=self.bc_lap)
        gradient = state.grid.get_operator('gradient', bc=self.bc)

        @jit(signature)
        def pde_rhs(state_data: np.ndarray, t: float):
            """ compiled helper function evaluating right hand side """ 
            grad = gradient(state_data)
            result = -laplace(state_data)
            result += nu_value * laplace2(result)
            for i in range(dim):
                result -= 0.5 * grad[i]**2
            return result
            
        return pde_rhs  # type: ignore
    
    
    



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import pytest
import numpy as np

from .. import PDE, SwiftHohenbergPDE
from ...grids import UnitGrid
from ...fields import ScalarField, VectorField, FieldCollection
from ...grids.tests.test_generic import iter_grids



def test_pde_wrong_input():
    """ test some wrong input """
    with pytest.raises(RuntimeError):
        PDE({'t': 1})
        
    grid = UnitGrid([4])
    eq = PDE({'u': 1})
    assert eq.expressions == {'u': '1.0'}
    with pytest.raises(ValueError):
        eq.evolution_rate(FieldCollection.scalar_random_uniform(2, grid))
    
    eq = PDE({'u': 1, 'v': 2})
    assert eq.expressions == {'u': '1.0', 'v': '2.0'}
    with pytest.raises(ValueError):
        eq.evolution_rate(ScalarField.random_uniform(grid))
        


def test_pde_scalar():
    """ test PDE with a single scalar field """
    eq = PDE({'u': 'laplace(u) + exp(-t) + sin(t)'})
    grid = UnitGrid([8])
    field = ScalarField.random_normal(grid)
    
    res_a = eq.solve(field, t_range=1, dt=0.01, backend='numpy', tracker=None)
    res_b = eq.solve(field, t_range=1, dt=0.01, backend='numba', tracker=None)
    
    res_a.assert_field_compatible(res_b)
    np.testing.assert_allclose(res_a.data, res_b.data)
    
    

def test_pde_vector():
    """ test PDE with a single vector field """
    eq = PDE({'u': 'vector_laplace(u) + exp(-t)'})
    grid = UnitGrid([8, 8])
    field = VectorField.random_normal(grid)

    res_a = eq.solve(field, t_range=1, dt=0.01, backend='numpy', tracker=None)
    res_b = eq.solve(field, t_range=1, dt=0.01, backend='numba', tracker=None)
    
    res_a.assert_field_compatible(res_b)
    np.testing.assert_allclose(res_a.data, res_b.data)    
    
    
    
def test_pde_2scalar():
    """ test PDE with two scalar fields """
    eq = PDE({'u': 'laplace(u) - u', 'v': '- u * v'})
    grid = UnitGrid([8])
    field = FieldCollection.scalar_random_uniform(2, grid)

    res_a = eq.solve(field, t_range=1, dt=0.01, backend='numpy', tracker=None)
    res_b = eq.solve(field, t_range=1, dt=0.01, backend='numba', tracker=None)
    
    res_a.assert_field_compatible(res_b)
    np.testing.assert_allclose(res_a.data, res_b.data)
    
    
    
def test_pde_vector_scalar():
    """ test PDE with a vector and a scalar field """
    eq = PDE({'u': 'vector_laplace(u) - u + gradient(v)',
              'v': '- divergence(u)'})
    grid = UnitGrid([8, 8])
    field = FieldCollection([VectorField.random_uniform(grid),
                             ScalarField.random_uniform(grid)])

    res_a = eq.solve(field, t_range=1, dt=0.01, backend='numpy', tracker=None)
    res_b = eq.solve(field, t_range=1, dt=0.01, backend='numba', tracker=None)
    
    res_a.assert_field_compatible(res_b)
    np.testing.assert_allclose(res_a.data, res_b.data)  



@pytest.mark.parametrize('grid', iter_grids())
def test_compare_swift_hohenberg(grid):
    """ compare custom class to swift-Hohenberg """
    rate, kc2, delta = np.random.uniform(0.5, 2, size=3)
    eq1 = SwiftHohenbergPDE(rate=rate, kc2=kc2, delta=delta)
    eq2 = PDE({'u': f"({rate} - {kc2}**2) * u - 2 * {kc2} * laplace(u) "
                    f"- laplace(laplace(u)) + {delta} * u**2 - u**3"})

    field = ScalarField.random_uniform(grid)
    res1 = eq1.solve(field, t_range=1, dt=0.01, backend='numpy', tracker=None)
    res2 = eq2.solve(field, t_range=1, dt=0.01, backend='numpy', tracker=None)
    
    res1.assert_field_compatible(res1)
    np.testing.assert_allclose(res1.data, res2.data)
    
    
    
def test_custom_operators():
    """ test using a custom operator """
    grid = UnitGrid([32])
    field = ScalarField.random_normal(grid)
    eq = PDE({'u': 'undefined(u)'})
    
    with pytest.raises(ValueError):
        eq.evolution_rate(field)
        
    
    def make_op(state):
        return lambda state: state
    UnitGrid.register_operator('undefined', make_op)
    
    eq._cache = {}  # reset cache
    res = eq.evolution_rate(field)
    np.testing.assert_allclose(field.data, res.data)
    
    del UnitGrid._operators['undefined']  # reset original state
    
    
    
@pytest.mark.parametrize('backend', ['numpy', 'numba'])
def test_pde_noise(backend):
    """ test noise operator on PDE class """
    grid = UnitGrid([64, 64])
    state = FieldCollection([ScalarField(grid), ScalarField(grid)])
    
    eq = PDE({'a': 0, 'b': 0}, noise=.5)
    res = eq.solve(state, t_range=1, backend=backend, dt=1)
    assert res.data.std() == pytest.approx(.5, rel=0.1)

    eq = PDE({'a': 0, 'b': 0}, noise=[0.01, 2.])
    res = eq.solve(state, t_range=1, backend=backend, dt=1)
    assert res.data[0].std() == pytest.approx(0.01, rel=0.1)
    assert res.data[1].std() == pytest.approx(2., rel=0.1)

    with pytest.raises(RuntimeError):
        eq = PDE({'a': 0}, noise=[0.01, 2.])
        eq.solve(ScalarField(grid), t_range=1, backend=backend, dt=1)
    


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np
import pytest

from ...grids import PolarGrid, SphericalGrid



@pytest.mark.parametrize("grid_class", [PolarGrid, SphericalGrid])
def test_spherical_base_bcs(grid_class):
    """ test setting boundary conditions on spherical grids """
    grid = grid_class(2, 3)
    
    domain1 = grid.get_boundary_conditions(['derivative', {'type': 'value'}])
    domain2 = grid.get_boundary_conditions({'type': 'value'})
    assert domain1 == domain2
    
    with pytest.raises(ValueError):
        grid.get_boundary_conditions(['value', 'value'])

    # test boundary conditions for simulations with holes
    grid = grid_class((1, 2), 3)
    grid.get_boundary_conditions(['derivative', {'type': 'value'}])
    domain1 = grid.get_boundary_conditions({'type': 'value'})
    domain2 = grid.get_boundary_conditions(['value', 'value'])
    assert domain1 == domain2



def test_polar_grid():
    """ test simple polar grid """
    grid = PolarGrid(4, 8)
    assert grid.dim == 2
    assert grid.numba_type == "f8[:]"
    assert grid.shape == (8,)
    assert not grid.has_hole
    assert grid.discretization[0] == pytest.approx(0.5)
    np.testing.assert_array_equal(grid.discretization, np.array([0.5]))
    assert grid.volume == pytest.approx(np.pi * 4**2)
    assert grid.volume == pytest.approx(grid.integrate(1))
    
    np.testing.assert_allclose(grid.axes_coords[0],
                               np.linspace(0.25, 3.75, 8))
    
    a = grid.get_operator('laplace', 'natural')(np.random.random(8))
    assert a.shape == (8,)
    assert np.all(np.isfinite(a))

    # random points
    c = np.random.randint(8, size=(6, 1))
    p = grid.cell_to_point(c)
    np.testing.assert_array_equal(c, grid.point_to_cell(p))

    assert grid.contains_point(grid.get_random_point())
    assert grid.contains_point(grid.get_random_point(3.99))
    assert "laplace" in grid.operators



def test_polar_annulus():
    """ test simple polar grid with a hole """
    grid = PolarGrid((2, 4), 8)
    assert grid.dim == 2
    assert grid.numba_type == "f8[:]"
    assert grid.shape == (8,)
    assert grid.has_hole
    assert grid.discretization[0] == pytest.approx(0.25)
    np.testing.assert_array_equal(grid.discretization, np.array([0.25]))
    assert grid.volume == pytest.approx(np.pi * (4**2 - 2**2))
    assert grid.volume == pytest.approx(grid.integrate(1))
    assert grid.radius == (2, 4)
    
    np.testing.assert_allclose(grid.axes_coords[0],
                               np.linspace(2.125, 3.875, 8))
    
    a = grid.get_operator('laplace', 'natural')(np.random.random(8))
    assert a.shape == (8,)
    assert np.all(np.isfinite(a))

    # random points
    c = np.random.randint(8, size=(6, 1))
    p = grid.cell_to_point(c)
    np.testing.assert_array_equal(c, grid.point_to_cell(p))

    assert grid.contains_point(grid.get_random_point())
    assert grid.contains_point(grid.get_random_point(1.99))
    
    # test boundary points
    np.testing.assert_equal(grid._boundary_coordinates(0, False), np.array([2]))
    np.testing.assert_equal(grid._boundary_coordinates(0, True), np.array([4]))



def test_polar_to_cartesian():
    """ test conversion of polar grid to Cartesian """
    from ...fields import ScalarField
    from .. import CartesianGrid
    
    expr_pol = '1 / (1 + r**2)'
    expr_cart = expr_pol.replace('r**2', '(x**2 + y**2)')
     
    grid_pol = PolarGrid(7, 16)
    pf_pol = ScalarField.from_expression(grid_pol, expression=expr_pol)
     
    grid_cart = CartesianGrid([[-4, 4], [-3.9, 4.1]], [16, 16])
    pf_cart1 = pf_pol.interpolate_to_grid(grid_cart)
    pf_cart2 = ScalarField.from_expression(grid_cart, expression=expr_cart)
    np.testing.assert_allclose(pf_cart1.data, pf_cart2.data, atol=0.1)
     


def test_spherical_grid():
    """ test simple spherical grid """
    grid = SphericalGrid(4, 8)
    assert grid.dim == 3
    assert grid.numba_type == "f8[:]"
    assert grid.shape == (8,)
    assert not grid.has_hole
    assert grid.discretization[0] == pytest.approx(0.5)
    np.testing.assert_array_equal(grid.discretization, np.array([0.5]))
    assert grid.volume == pytest.approx(4/3*np.pi * 4**3)
    assert grid.volume == pytest.approx(grid.integrate(1))
    
    np.testing.assert_allclose(grid.axes_coords[0],
                               np.linspace(0.25, 3.75, 8))
    
    a = grid.get_operator('laplace', 'natural')(np.random.random(8))
    assert a.shape == (8,)
    assert np.all(np.isfinite(a))

    # random points
    c = np.random.randint(8, size=(6, 1))
    p = grid.cell_to_point(c)
    np.testing.assert_array_equal(c, grid.point_to_cell(p))

    assert grid.contains_point(grid.get_random_point())
    assert grid.contains_point(grid.get_random_point(3.99))
    assert "laplace" in grid.operators



def test_spherical_annulus():
    """ test simple spherical grid with a hole """
    grid = SphericalGrid((2, 4), 8)
    assert grid.dim == 3
    assert grid.numba_type == "f8[:]"
    assert grid.shape == (8,)
    assert grid.has_hole
    assert grid.discretization[0] == pytest.approx(0.25)
    np.testing.assert_array_equal(grid.discretization, np.array([0.25]))
    assert grid.volume == pytest.approx(4/3*np.pi * (4**3 - 2**3))
    assert grid.volume == pytest.approx(grid.integrate(1))
    assert grid.radius == (2, 4)
    
    np.testing.assert_allclose(grid.axes_coords[0],
                               np.linspace(2.125, 3.875, 8))
    
    a = grid.get_operator('laplace', 'natural')(np.random.random(8))
    assert a.shape == (8,)
    assert np.all(np.isfinite(a))

    # random points
    c = np.random.randint(8, size=(6, 1))
    p = grid.cell_to_point(c)
    np.testing.assert_array_equal(c, grid.point_to_cell(p))

    assert grid.contains_point(grid.get_random_point())
    assert grid.contains_point(grid.get_random_point(1.99))

    # test boundary points
    np.testing.assert_equal(grid._boundary_coordinates(0, False), np.array([2]))
    np.testing.assert_equal(grid._boundary_coordinates(0, True), np.array([4]))

 
 
def test_spherical_to_cartesian():
    """ test conversion of spherical grid to cartesian """
    from ...fields import ScalarField
    from .. import CartesianGrid
    
    expr_sph = '1 / (1 + r**2)'
    expr_cart = expr_sph.replace('r**2', '(x**2 + y**2 + z**2)')
    
    grid_sph = SphericalGrid(7, 16)
    pf_sph = ScalarField.from_expression(grid_sph, expression=expr_sph)
     
    grid_cart = CartesianGrid([[-4, 4], [-3.9, 4.1], [-4.1, 3.9]], [16] * 3)
    pf_cart1 = pf_sph.interpolate_to_grid(grid_cart)
    pf_cart2 = ScalarField.from_expression(grid_cart, expression=expr_cart)
    np.testing.assert_allclose(pf_cart1.data, pf_cart2.data, atol=0.1)



#!/usr/bin/env python3

import sys
from pathlib import Path

PACKAGE_PATH = Path(__file__).resolve().parents[1]
sys.path.append(str(PACKAGE_PATH))

import numpy as np
import numba

from pde.grids import UnitGrid, CylindricalGrid, SphericalGrid
from pde.grids.operators import cartesian, cylindrical, spherical
from pde.grids.operators.common import PARALLELIZATION_THRESHOLD_2D
from pde.grids.boundaries import Boundaries
from pde.tools.numba import jit, jit_allocate_out
from pde.tools.misc import estimate_computation_speed



def custom_laplace_2d_periodic(shape, dx=1):
    """ make laplace operator with periodic boundary conditions """
    dx_2 = 1 / dx**2  
    dim_x, dim_y = shape    
    parallel = (dim_x * dim_y >= PARALLELIZATION_THRESHOLD_2D**2)
    
    @jit_allocate_out(parallel=parallel)
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        for i in numba.prange(dim_x):
            im = dim_x - 1 if i == 0 else i - 1
            ip = 0 if i == dim_x - 1 else i + 1
                
            j = 0
            jm = dim_y - 1
            jp = j + 1
            out[i, j] = (arr[i, jm] + arr[i, jp] + arr[im, j] + arr[ip, j] -
                         4 * arr[i, j]) * dx_2
                         
            for j in range(1, dim_y - 1):
                jm = j - 1
                jp = j + 1
                out[i, j] = (arr[i, jm] + arr[i, jp] + arr[im, j] + arr[ip, j] -
                             4 * arr[i, j]) * dx_2
                             
            j = dim_y - 1
            jm = j - 1
            jp = 0
            out[i, j] = (arr[i, jm] + arr[i, jp] + arr[im, j] + arr[ip, j] -
                         4 * arr[i, j]) * dx_2
        return out
    
    return laplace



def custom_laplace_2d_no_flux(shape, dx=1):
    """ make laplace operator with no-flux boundary conditions """
    dx_2 = 1 / dx**2  
    dim_x, dim_y = shape
    parallel = (dim_x * dim_y >= PARALLELIZATION_THRESHOLD_2D**2)

    @jit_allocate_out(parallel=parallel)
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        for i in numba.prange(dim_x):
            im = 0 if i == 0 else i - 1
            ip = dim_x - 1 if i == dim_x - 1 else i + 1
                
            for j in range(dim_y):
                jm = 0 if j == 0 else j - 1
                jp = dim_y - 1 if j == dim_y - 1 else j + 1
                    
                out[i, j] = (arr[i, jm] + arr[i, jp] +
                             arr[im, j] + arr[ip, j] -
                             4 * arr[i, j]) * dx_2
        return out
    return laplace



def custom_laplace_2d(shape, periodic, dx=1):
    """ make laplace operator with no-flux or periodic boundary conditions """
    if periodic:
        return custom_laplace_2d_periodic(shape, dx=dx)
    else:
        return custom_laplace_2d_no_flux(shape, dx=dx)



def flexible_laplace_2d(bcs):
    """ make laplace operator with flexible boundary conditions """
    bc_x, bc_y = bcs
    dx = bcs._uniform_discretization
    dx_2 = 1 / dx**2
    dim_x, dim_y = bcs.grid.shape
     
    region_x = bc_x.make_region_evaluator()
    region_y = bc_y.make_region_evaluator()
     
    @jit_allocate_out
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        for i in range(dim_x):
            for j in range(dim_y):
                val_x_l, val_x, val_x_r = region_x(arr, (i, j))
                val_y_l, _, val_y_r = region_y(arr, (i, j))

                out[i, j] = (val_x_l + val_x_r + val_y_l + val_y_r -
                             4 * val_x) * dx_2
        return out        
     
    return laplace



def custom_laplace_cyl_no_flux(shape, dr=1, dz=1):
    """ make laplace operator with no-flux boundary conditions """
    dim_r, dim_z = shape
    dr_2 = 1 / dr**2
    dz_2 = 1 / dz**2

    @jit
    def laplace(arr, out=None):
        """ apply laplace operator to array `arr` """
        if out is None:
            out = np.empty((dim_r, dim_z))
            
        for j in range(0, dim_z):  # iterate axial points
            jm = 0 if j == 0 else j - 1
            jp = dim_z - 1 if j == dim_z - 1 else j + 1
                
            # inner radial boundary condition
            i = 0
            out[i, j] = (
                2 * (arr[i + 1, j] - arr[i, j]) * dr_2 +
                (arr[i, jm] + arr[i, jp] - 2 * arr[i, j]) * dz_2
            )
            
            for i in range(1, dim_r - 1):  # iterate radial points
                out[i, j] = (
                    (arr[i + 1, j] - 2 * arr[i, j] + arr[i - 1, j]) * dr_2 +
                    (arr[i + 1, j] - arr[i - 1, j]) / (2 * i + 1) * dr_2 +
                    (arr[i, jm] + arr[i, jp] - 2 * arr[i, j]) * dz_2
                )
                
            # outer radial boundary condition
            i = dim_r - 1
            out[i, j] = (
                (arr[i - 1, j] - arr[i, j]) * dr_2 +
                (arr[i, j] - arr[i - 1, j]) / (2 * i + 1) * dr_2 +
                (arr[i, jm] + arr[i, jp] - 2 * arr[i, j]) * dz_2
            )
        return out
        
    return laplace



def main():
    """ main routine testing the performance """
    print('Reports calls-per-second (larger is better)')
    print('  The `CUSTOM` method implemented by hand is the baseline case.')
    print('  The `FLEXIBLE` method is a serial implementation using the '
          'boundary conditions from the package.')
    print('  The other methods use the functions supplied by the package.\n')
    
    # Cartesian grid with different shapes and boundary conditions
    for shape in [(32, 32), (512, 512)]:
        data = np.random.random(shape)
        for periodic in [True, False]:
            grid = UnitGrid(shape, periodic=periodic)
            bcs = Boundaries.from_data(grid, 'natural')
            print(grid)
            result = cartesian.make_laplace(bcs, method='scipy')(data)
             
            for method in ['FLEXIBLE', 'CUSTOM', 'numba', 'matrix', 'scipy']:
                if method == 'FLEXIBLE':
                    laplace = flexible_laplace_2d(bcs)
                elif method == 'CUSTOM':
                    laplace = custom_laplace_2d(shape, periodic=periodic)
                elif method in {'numba', 'matrix', 'scipy'}:
                    laplace = cartesian.make_laplace(bcs, method=method)
                else:
                    raise ValueError(f'Unknown method `{method}`')
                # call once to pre-compile and test result 
                np.testing.assert_allclose(laplace(data), result)
                speed = estimate_computation_speed(laplace, data)
                print(f'{method:>8s}: {int(speed):>9d}')
            print()
 
    # Cylindrical grid with different shapes
    for shape in [(32, 64), (512, 512)]:
        data = np.random.random(shape)
        grid = CylindricalGrid(shape[0], [0, shape[1]], shape)
        print(f'Cylindrical grid, shape={shape}')
        bcs = Boundaries.from_data(grid, 'no-flux')
        laplace_cyl = cylindrical.make_laplace(bcs)
        result = laplace_cyl(data)
         
        for method in ['CUSTOM', 'numba']:
            if method == 'CUSTOM':
                laplace = custom_laplace_cyl_no_flux(shape)
            elif method == 'numba':
                laplace = laplace_cyl
            else:
                raise ValueError(f'Unknown method `{method}`')
            # call once to pre-compile and test result 
            np.testing.assert_allclose(laplace(data), result)
            speed = estimate_computation_speed(laplace, data)
            print(f'{method:>8s}: {int(speed):>9d}')
        print()

    # Spherical grid with different shapes
    for shape in [32, 512]:
        data = np.random.random(shape)
        grid = SphericalGrid(shape, shape)
        print(grid)
        bcs = Boundaries.from_data(grid, 'no-flux')        
        make_laplace = spherical.make_laplace 
        
        for conservative in [True, False]:
            laplace = make_laplace(bcs, conservative=conservative)
            # call once to pre-compile 
            laplace(data)
            speed = estimate_computation_speed(laplace, data)
            print(f' numba (conservative={str(conservative):<5}): '
                  f'{int(speed):>9d}')
        print()



if __name__ == '__main__':
    main()
